#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/gfp.h>
#include <linux/idr.h>
#include <linux/slab.h>
#include "gt_debug.h"
#include "gt_misc.h"
#include "gt_idr.h"

gt_idr *gt_idr_kalloc(void)
{
	struct idr *gtidr = (struct idr *)kzalloc(sizeof(struct idr), GFP_KERNEL);
	if (!gtidr) {
		gt_error("[%s: %d] kzalloc %d bytes failed\n", __func__, __LINE__, sizeof(struct idr));
		return NULL;
	}

	idr_init(gtidr);

	return gtidr;
}
GT_EXT_SYM(gt_idr_kalloc);

void gt_idr_init(gt_idr *gtidr)
{
	idr_init((struct idr *)gtidr);
}
GT_EXT_SYM(gt_idr_init);

void gt_idr_free(gt_idr *gtidr)
{
	kfree((struct idr*)gtidr);
}
GT_EXT_SYM(gt_idr_free);

int gt_idr_alloc(gt_idr *gtidr, void *ptr, int start, int end, int *id)
{
	int result;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0))
	idr_preload(GFP_KERNEL);
	*id=idr_alloc((struct idr*)gtidr, ptr, start, end + 1,0);
	idr_preload_end();

	result = *id;
#else
	do
	{
		if(idr_pre_get((struct idr*)gtidr, GFP_KERNEL)==0)
		{
			return -1;
		}
		result = idr_get_new_above((struct idr*)gtidr, ptr, start, *id);
	}while(result == -EAGAIN);

	if ((IMG_UINT32)*id > end)
	{
		idr_remove((struct idr*)gtidr, *id);
		result = -EAGAIN;
	}
#endif

	return result;
}
GT_EXT_SYM(gt_idr_alloc);

int gt_idr_alloc_cyclic(gt_idr *gtidr, void *entry, int start, int end)
{
	return idr_alloc_cyclic((struct idr*)gtidr, entry, start, end, GFP_KERNEL);
}
GT_EXT_SYM(gt_idr_alloc_cyclic);

void * gt_idr_get_next(gt_idr *gtidr, int *nextid)
{
	return idr_get_next((struct idr*)gtidr, nextid);
}
GT_EXT_SYM(gt_idr_get_next);

void * gt_idr_find(gt_idr *gtidr, int id)
{
	return idr_find((struct idr*)gtidr, id);
}
GT_EXT_SYM(gt_idr_find);

void * gt_idr_remove(gt_idr *gtidr, int id)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	idr_remove((struct idr*)gtidr, id);
	return NULL;
#else
	return idr_remove((struct idr*)gtidr, id);
#endif
}
GT_EXT_SYM(gt_idr_remove);

void gt_idr_destroy(gt_idr *gtidr)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0))
	idr_remove_all((struct idr*)gtidr);
#endif
	idr_destroy((struct idr*)gtidr);
}
GT_EXT_SYM(gt_idr_destroy);

int gt_idr_foreach(gt_idr *gtidr,
				int (*function)(int id, void *p, void *data), void *data)
{
	return idr_for_each((struct idr*)gtidr, function, data);
}
GT_EXT_SYM(gt_idr_foreach);

void *gt_idr_replace(gt_idr *gtidr, void *pvData, int id)
{
	return idr_replace((struct idr*)gtidr, pvData, id);
}
GT_EXT_SYM(gt_idr_replace);
