/*
 * Copyright (c) Gengtu Technology Ltd. All Rights Reserved
 * Dual MIT/GPLv2
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License Version 2 ("GPL") in which case the provisions
 * of GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms of
 * GPL, and not to allow others to use your version of this file under the terms
 * of the MIT license, indicate your decision by deleting the provisions above
 * and replace them with the notice and other provisions required by GPL as set
 * out in the file called "GPL-COPYING" included in this distribution. If you do
 * not delete the provisions above, a recipient may use your version of this file
 * under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT-COPYING".
 *
 * EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
 * PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <linux/platform_device.h>
#include <linux/version.h>
#include <linux/memblock.h>
#include <linux/kallsyms.h>
#include <linux/iommu.h>
#include "gt_misc.h"
#include "gt_plat_dev.h"
#include <linux/vmalloc.h>
#include <linux/device.h>
#include <linux/numa.h>

gt_platform_device *gt_platform_device_register_full(
		struct gt_plat_dev_info *_info)
{
	gt_platform_device *ret = NULL;
	struct resource *res = NULL;
	uint32_t num_res = 0;
	uint32_t i = 0;

	struct platform_device_info info = {
		.parent    = _info->parent,
		.name      = _info->name,
		.id        = _info->id,
		.data      = _info->data,
		.size_data = _info->data_size,
		.dma_mask  = _info->dma_mask,
	};

	if (_info->res) {
		if (_info->num_res == 0)
			num_res = 1;
		else 
			num_res = _info->num_res;

		res = (struct resource *)vmalloc(num_res * sizeof(struct resource));
		if (!res) {
			pr_err("no memory!\n");
			return NULL;
		}
		for (i = 0; i < num_res; i++) {
			struct resource res_local = DEFINE_RES_MEM_NAMED(_info->res->start, _info->res->size, _info->res->name);
			res[i] = res_local;
        }

		info.res     = res;
		info.num_res = num_res;
	}

	ret = platform_device_register_full(&info);
	if (_info->res) {
		vfree(res);
	}

	return ret;
}
GT_EXT_SYM(gt_platform_device_register_full);

void gt_platform_device_unregister(gt_platform_device *dev)
{
	platform_device_unregister((struct platform_device *)dev);
}
GT_EXT_SYM(gt_platform_device_unregister);

void *gt_platform_get_parent_dev(gt_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return pdev->dev.parent;
}
GT_EXT_SYM(gt_platform_get_parent_dev);

void *gt_platform_get_dev(gt_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return &pdev->dev;
}
GT_EXT_SYM(gt_platform_get_dev);

gt_dev *gt_dev_get_parent(gt_dev *dev)
{
	return ((struct device *)dev)->parent;
}
GT_EXT_SYM(gt_dev_get_parent);

gt_platform_device *gt_to_platform_device(void *dev)
{
	struct device *_dev = (struct device *)dev;
	return to_platform_device(_dev);
}
GT_EXT_SYM(gt_to_platform_device);

void *gt_platform_get_data(gt_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return pdev->dev.platform_data;
}
GT_EXT_SYM(gt_platform_get_data);

uint64_t gt_platform_get_dma_mask(gt_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return *pdev->dev.dma_mask;
}
GT_EXT_SYM(gt_platform_get_dma_mask);

gt_resource *gt_platform_get_resource_byname(gt_platform_device *dev,
		unsigned int type, const char *name)
{
	return platform_get_resource_byname((struct platform_device *)dev,
		type, name);
}
GT_EXT_SYM(gt_platform_get_resource_byname);

uint64_t gt_resource_start(gt_resource *res)
{
	return ((struct resource *)res)->start;
}
GT_EXT_SYM(gt_resource_start);

uint64_t gt_resource_size(gt_resource *res)
{
	return resource_size((struct resource *)res);
}
GT_EXT_SYM(gt_resource_size);

static void gt_rsrc_devres_release(struct device *dev, void *res)
{
	/* No extra cleanup needed */
}

void *gt_rsrc_devres_find(gt_dev *dev)
{
	return devres_find(dev, gt_rsrc_devres_release, NULL, NULL);
}
GT_EXT_SYM(gt_rsrc_devres_find);

void *gt_rsrc_devres_alloc(size_t size)
{
	return devres_alloc(gt_rsrc_devres_release, size, GFP_KERNEL);
}

struct device *gt_dev_to_dev(gt_dev *dev)
{
	return (struct device *)dev;
}
GT_EXT_SYM(gt_dev_to_dev);

void gt_put_device(gt_dev *dev)
{
	put_device((struct device *)dev);
}
GT_EXT_SYM(gt_put_device);

gt_dev *gt_get_device(gt_dev *dev)
{
	return (gt_dev *)get_device((struct device *)dev);
}
GT_EXT_SYM(gt_get_device);

void gt_platform_set_drvdata(void *pdev, void *data)
{
	platform_set_drvdata((struct platform_device *)pdev, data);
}
GT_EXT_SYM(gt_platform_set_drvdata);

void *gt_platform_get_drvdata(void *pdev)
{
	return platform_get_drvdata((struct platform_device *)pdev);
}
GT_EXT_SYM(gt_platform_get_drvdata);

void *gt_dev_get_platdata(void *pdev)
{
	return (void *)dev_get_platdata((struct device *)pdev);
}
GT_EXT_SYM(gt_dev_get_platdata);

void *gt_get_dev_ofnode(gt_dev *dev)
{
	return ((struct device *)dev)->of_node;
}
GT_EXT_SYM(gt_get_dev_ofnode);

void *gt_get_platdev_drvdata(void *pdev)
{
	return dev_get_drvdata(&((struct platform_device *)pdev)->dev);
}
GT_EXT_SYM(gt_get_platdev_drvdata);

void *gt_get_dev_drvdata(gt_dev *dev)
{
	return dev_get_drvdata((struct device *)dev);
}
GT_EXT_SYM(gt_get_dev_drvdata);

void gt_set_dev_drvdata(gt_dev *dev, void *data)
{
	dev_set_drvdata((struct device *)dev, data);
}
GT_EXT_SYM(gt_set_dev_drvdata);

int gt_platform_get_numa_node(gt_platform_device *dev)
{
	int node_node = 0;
#ifdef CONFIG_NUMA
	struct platform_device *pdev = (struct platform_device *)dev;
	node_node = pdev->dev.numa_node;
#endif
	return node_node;
}
GT_EXT_SYM(gt_platform_get_numa_node);

int gt_dev_to_node(gt_dev *dev)
{
	return dev_to_node(dev);
}
GT_EXT_SYM(gt_dev_to_node);

void gt_platform_suspend(gt_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	if (pdev &&
			pdev->dev.driver &&
			pdev->dev.driver->pm &&
			pdev->dev.driver->pm->suspend) {
		pdev->dev.driver->pm->suspend(&pdev->dev);
	}
}
GT_EXT_SYM(gt_platform_suspend);

bool gt_get_iommu_enable(gt_dev* dev)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	struct bus_type *bus = ((struct device *)dev)->bus;
	if (iommu_present(bus)) {
		return true;
	}
#else
#ifdef CONFIG_DMA_OPS
	if (((struct device *)dev)->dma_ops != NULL) {
		return true;
	}
#endif
#endif

	return false;
}
GT_EXT_SYM(gt_get_iommu_enable);

