#ifdef CONFIG_PM_DEVFREQ
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/devfreq.h>
#include <linux/math64.h>
#include <linux/devfreq.h>
#include <linux/thermal.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>

#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/poll.h>

#include "gt_devfreq_gov.h"
#include "gt_input_event.h"
#include "hal.h"
#include "gt_task.h"
#include "gt_misc.h"
#include "gt_debug.h"

#define GTGOV "dfreqgov"
#define gtgov(fmt) "[%s][%s:%d]" fmt,GTGOV,__func__,__LINE__
#define gtgov_info(fmt, ...) \
	if (get_pwr_debug_lvl() >= PWRD_LVL_INFO) \
		gt_printk(KERN_INFO gtgov(fmt), ##__VA_ARGS__)

#define gtgov_dbg(fmt, ...) \
	if (get_pwr_debug_lvl() >= PWRD_LVL_DBG) \
		gt_printk(KERN_NOTICE gtgov(fmt), ##__VA_ARGS__)

#define gtgov_notice( fmt, ...) \
	gt_printk(KERN_NOTICE gtgov(fmt), ##__VA_ARGS__)

#define gtgov_warn( fmt, ...) \
	gt_printk(KERN_WARNING gtgov(fmt), ##__VA_ARGS__)

#define gtgov_err(fmt, ...) \
	gt_printk(KERN_NOTICE gtgov(fmt), ##__VA_ARGS__)

#define GTIDLE_ENTER_IDLE_THR 10
#define GTIDLE_EXIT_IDLE_THR  20

#define GTIDLE_DETECT_PERIOD_TIME 1
#define GTIDLE_ENTER_IDLE_CONTINUOUS_DETECT_TIME 30
#define GTIDLE_EXIT_IDLE_CONTINUOUS_DETECT_TIME 5

#define GT_FAILED -1
#define GT_SUCC   0

static struct devfreq_info gdevfreq_info = {
	false, NULL, NULL
};

static int gt_devfreq_bind_input_handler(void *osdev, void *devfreq)
{
	gdevfreq_info.devfreq = devfreq;
	gdevfreq_info.osdev = osdev;
	gdevfreq_info.inited = true;

	return 0;
}

static int gt_devfreq_unbind_input_handler(void)
{
	gdevfreq_info.inited = false;
	gdevfreq_info.devfreq = NULL;
	gdevfreq_info.osdev = NULL;

	return 0;
}

struct devfreq_info* gt_get_devfreq_info(void)
{
	return &gdevfreq_info;
}

/* called with reporting input events and reporting is interrupt context, so need to consider schedule */
void gt_input_kick(void *dev)
{
	struct devfreq *devfreq = NULL;
	struct devfreq_gt_govdata *gtgov_data = NULL;
	struct devfreq_info *pdevfreq_info = (struct devfreq_info *)gt_get_devfreq_info();

	if (!pdevfreq_info->inited) {
		gtgov_info("devfreq info is not inited\n");
		return;
	}

	devfreq = pdevfreq_info->devfreq;
	gtgov_data = devfreq->data;

	gtgov_data->idle_enter_cnt = 0;
	gtgov_data->idle_exit_cnt = 0;
	if (get_pwr_debug_lvl() == PWRD_DBG_INPUT) {
		gtgov_notice("dev is %p devfreq is %p gtgov_data is %p load state is %s\n", dev, devfreq, gtgov_data,
		gtgov_data->load_state == LOAD_STATE_IDLE ? "LOAD_STATE_IDLE" : "LOAD_STATE_ACTIVE");
	}
	if (gtgov_data->load_state == LOAD_STATE_IDLE) {		//load state is changed
		gtgov_dbg("load state is changed: idle ===> active\n");
		gtgov_data->load_state = LOAD_STATE_ACTIVE;
		gtgov_data->input_kicked = true;
		if (!gt_queue_dwork(gtgov_data->devfreq_wkq, gtgov_data->input_kick_dwork, 0)) {
			gtgov_notice("failed to start input kick dwork\n");
		}
	}

	return ;
}

static int gt_idle_check(struct devfreq *devfreq, unsigned long *freq)
{
	int util = 0;
	bool load_changed = false, input_kicked = false;
	int idle_exit_cnt = 0, idle_enter_cnt = 0;
	struct devfreq_gt_govdata *govdata = devfreq->data;
	static unsigned long pre_freq = 0;

	util = hal_get_gpu_utils(govdata->gt69srv_dev_node);
	if (util < 0 && util > 100) {
		govdata->idle_enter_cnt = 0;
		govdata->idle_exit_cnt = 0;
		gtgov_err("util get failed util is %d\n", util);
		return GT_FAILED;
	}

	if (govdata->load_state == LOAD_STATE_ACTIVE) { 	//LOAD_STATE_ACTIVE --> LOAD_STATE_IDLE: <=10
		if (util >=0 && util <= GTIDLE_ENTER_IDLE_THR) {
			govdata->idle_exit_cnt = 0;
			govdata->idle_enter_cnt++;
			if (govdata->idle_enter_cnt >= GTIDLE_ENTER_IDLE_CONTINUOUS_DETECT_TIME) {
				govdata->load_state = LOAD_STATE_IDLE;
				*freq = govdata->user_powsave_freq + govdata->freq_offset;
				pre_freq = *freq;
				govdata->user_mode_powsave = true;

				load_changed = true;
			}
		} else {
			govdata->idle_enter_cnt = 0;
		}
	} else {
		if (util > GTIDLE_EXIT_IDLE_THR) {
			govdata->idle_enter_cnt = 0;
			govdata->idle_exit_cnt++;
			if (govdata->idle_exit_cnt >= GTIDLE_EXIT_IDLE_CONTINUOUS_DETECT_TIME) {
				govdata->load_state = LOAD_STATE_ACTIVE;
				*freq = govdata->default_freq + govdata->freq_offset;
				pre_freq = *freq;
				govdata->user_mode_powsave = false;

				load_changed = true;
			}
		} else {
			govdata->idle_exit_cnt = 0;
		}
	}

	if (!load_changed) {
		*freq = pre_freq;
	}

	/*clear cnt and kick data*/
	idle_exit_cnt = govdata->idle_exit_cnt;
	idle_enter_cnt = govdata->idle_enter_cnt;
	if (load_changed) {
		govdata->idle_exit_cnt = 0;
		govdata->idle_enter_cnt = 0;
	}

	if (govdata->input_kicked) {
		input_kicked = govdata->input_kicked;
		govdata->user_mode_powsave = false;
		govdata->load_state = LOAD_STATE_ACTIVE;
		*freq = govdata->default_freq + govdata->freq_offset;
		pre_freq = *freq;
		govdata->input_kicked = false;
	}

	gtgov_dbg("govdata 0x%p util %d load[%s] idle_enter_cnt[%d %d] idle_exit_cnt[%d %d] " \
		"pre_freq = %d *freq = %d input_kicked[%d %d] load_changed[%d]\n", govdata,
		util, govdata->load_state == LOAD_STATE_IDLE ? "LOAD_STATE_IDLE" : "LOAD_STATE_ACTIVE",
		idle_enter_cnt, govdata->idle_enter_cnt, idle_exit_cnt, govdata->idle_exit_cnt, pre_freq,
		*freq, input_kicked, govdata->input_kicked, load_changed);
	return GT_SUCC;
}

static void gt_input_kick_work(void *df)
{
	int err = 0;
	struct devfreq* devfreq = (struct devfreq* )df;

	mutex_lock(&devfreq->lock);
	err = update_devfreq(devfreq);
	if (err)
		gtgov_notice("dvfs failed with (%d) error\n", err);
	mutex_unlock(&devfreq->lock);
}

unsigned long devfreq_update_algo_pid(struct devfreq *dev)
{
	struct devfreq_gt_govdata *gtgov_data = dev->data;
	struct gtpower_temp_pidctl *pidc = &(gtgov_data->pidc);
	struct gtthermal *pgtthermal = gtgov_data->thermal;
	struct thermal_zone_device *tz = pgtthermal->thermal_dev;

	long target_freq = 0, delta_freq = 0;
	long tolerance_win_min = 0, tolerance_win_max = 0;
	int i = 0, cur_temp = 0, cur_tmp_error = 0;

	if (!tz || !tz->ops || !tz->ops->get_temp) {
		gtgov_notice("thermal_dev is null\n");
		return 0;
	}

	tz->ops->get_temp(tz, &cur_temp);
	pidc->time++;
	if (pidc->time < pidc->timeout) {
		target_freq = pidc->real_target_freq;
		goto tget_skip;
	} 
	pidc->time = 0;

	if (cur_temp < pidc->recover_temp) {
		cur_tmp_error = 0;
		pidc->pre_tmp_error = 0;
		pidc->ppre_tmp_error = 0;
		target_freq = pidc->max_freq;
		pidc->real_target_freq = target_freq;
		pidc->pid_out_freq = target_freq;
		goto tget_freq;
	}
	if (cur_temp > pidc->warnning_temp) {
		cur_tmp_error = 0;
		pidc->pre_tmp_error = 0;
		pidc->ppre_tmp_error = 0;
		target_freq = pidc->min_freq;
		pidc->real_target_freq = target_freq;
		pidc->pid_out_freq = target_freq;
		goto tget_freq;
	}

	cur_tmp_error = pidc->ref_temp - cur_temp;
	/*algo: u(k) - u(k-1) = P*[e(k) - e(k-1)] + I*e(k) + D*[e(k) - 2e(k-1) + e(k-2)] */
	delta_freq = pidc->P * (cur_tmp_error - pidc->pre_tmp_error) +
				pidc->I * cur_tmp_error +
				pidc->D * (cur_tmp_error - 2 * pidc->pre_tmp_error + pidc->ppre_tmp_error);

	/*u(k) = u(k-1) + delta*/
	target_freq = pidc->pid_out_freq + delta_freq * GPU_FREQ_M2HZ_UNIT;
	pidc->pid_out_freq = target_freq; 

	/*make sure target freq is valid and avoid error is too big by parame  PID's I */
	if (target_freq <= pidc->min_freq) {
		target_freq = pidc->min_freq;
		pidc->pid_out_freq = target_freq; 
	} else if (target_freq >= pidc->max_freq) {
		target_freq = pidc->max_freq;
		pidc->pid_out_freq = target_freq; 
	} else {
		for (i = 0; i < PIDCTL_OUT_FREQ_NUM-1; i++) {
			if (target_freq == pidc->outfreq[i]) {
				goto tget_freq;
			}

			if (target_freq >= pidc->outfreq[i+1] && 
				target_freq < pidc->outfreq[i]) {

				tolerance_win_min = (pidc->outfreq[i] + pidc->outfreq[i+1])/2 - pidc->tolerance;
				tolerance_win_max = (pidc->outfreq[i] + pidc->outfreq[i+1])/2 + pidc->tolerance;

				/*tolerance adjust*/
				if (target_freq >= tolerance_win_min && target_freq <= tolerance_win_max) {
					target_freq = pidc->real_target_freq;
					goto tget_freq;
				}

				if (target_freq < tolerance_win_min) {
					target_freq = pidc->outfreq[i+1];
				} else {
					target_freq = pidc->outfreq[i];
				}
				goto tget_freq;
			}
		}
	}

tget_freq:
	if (get_pwr_debug_lvl() >= PWRD_LVL_INFO) {
		gtgov_info("time = %d temp=[%d, %d] err=[%d, %d, %d] PID=[%d, %d, %d]" \
			" delta_freq[%ld] target_freq=[pid:%ld, real_target%ld pre_real_target=%ld] min~max freq=[%ld, %ld]\n",
			pidc->time, cur_temp, pidc->ref_temp,
			cur_tmp_error, pidc->pre_tmp_error, pidc->ppre_tmp_error,
			pidc->P, pidc->I, pidc->D,
			delta_freq*GPU_FREQ_M2HZ_UNIT, pidc->pid_out_freq, target_freq, pidc->real_target_freq,
			pidc->min_freq, pidc->max_freq);
	}

	pidc->ppre_tmp_error = pidc->pre_tmp_error;
	pidc->pre_tmp_error = cur_tmp_error;
	pidc->real_target_freq = target_freq;
tget_skip:
	if (get_pwr_debug_lvl() >= PWRD_LVL_INFO) {
		gtgov_info("time=%d, cur_temp, pid_out_freq, real_target_freq=[%d, %ld, %ld]",
			pidc->time, cur_temp, pidc->pid_out_freq, pidc->real_target_freq);
	}

	return (unsigned long)target_freq;
}

#define CHIP_TEMP_DROP_THD		85
#define CHIP_TEMP_RECOVER_THD	(CHIP_TEMP_DROP_THD-5)
#define BOARD_TEMP_DROP_THD		80
#define BOARD_TEMP_RECOVER_THD	(BOARD_TEMP_DROP_THD-5)
#define KEEP_DROPFREQ_STAT	1
#define EXIT_DROPFREQ_STAT	0
unsigned long devfreq_update_algo_mix(struct devfreq *dev)
{
	struct devfreq_gt_govdata *gtgov_data = dev->data;
	struct gtthermal *pgtthermal = gtgov_data->thermal;
	struct thermal_zone_device *tz = pgtthermal->thermal_dev;
	struct thermal_zone_device *btz = pgtthermal->thermal_bdev;

	int chip_temp = 0;
	int board_temp = 0;
	long target_freq = 0;

	if (!tz || !tz->ops || !tz->ops->get_temp) {
		gtgov_notice("chip thermal_dev is null\n");
		return 0;
	}

	if (!btz || !btz->ops || !btz->ops->get_temp) {
		gtgov_notice("board thermal_dev is null\n");
		return 0;
	}

	tz->ops->get_temp(tz, &chip_temp);
	btz->ops->get_temp(btz, &board_temp);

	if (gtgov_data->keep_dropfreq_stat == KEEP_DROPFREQ_STAT) {
		target_freq = hal_get_gpu_drop_freq(pgtthermal->ppci_bdev) * GPU_FREQ_M2HZ_UNIT;			/*400M*/
	} else {
		target_freq = hal_get_gpu_recover_freq(pgtthermal->ppci_bdev) * GPU_FREQ_M2HZ_UNIT;		/*1000M*/
	}

	if (chip_temp >= CHIP_TEMP_DROP_THD || board_temp >= BOARD_TEMP_DROP_THD) {
		target_freq = hal_get_gpu_drop_freq(pgtthermal->ppci_bdev) * GPU_FREQ_M2HZ_UNIT;
		gtgov_data->keep_dropfreq_stat = KEEP_DROPFREQ_STAT;
	} else {
		if (chip_temp < CHIP_TEMP_RECOVER_THD && board_temp < BOARD_TEMP_RECOVER_THD) {
			target_freq = hal_get_gpu_recover_freq(pgtthermal->ppci_bdev) * GPU_FREQ_M2HZ_UNIT;
			gtgov_data->keep_dropfreq_stat = EXIT_DROPFREQ_STAT;
		}
	}
	gtgov_info("devfreq--%d-- chip_temp = %d board_temp = %d target_freq = %ld\n\n",gtgov_data->id, chip_temp, board_temp, target_freq);

	return target_freq;
}

unsigned long devfreq_update_by_thermal(struct devfreq *dev)
{
	struct devfreq_gt_govdata *gov_data = dev->data;
	struct gtthermal *pgtthermal = gov_data->thermal;
	int dynfreq_algo = 0;

	if (gt_is_err_or_null(pgtthermal)) {
		gtgov_notice("pgtthermal is null\n");
		return 0;
	}

	if (gt_is_err_or_null(pgtthermal->ppci_bdev)) {
		gtgov_notice("pgtthermal->ppci_bdev is null\n");
		return 0;
	}

	dynfreq_algo = hal_get_dynfreq_algo(pgtthermal->ppci_bdev);
	if (dynfreq_algo == DYNFREQ_ALGO_MIX) {
		return devfreq_update_algo_mix(dev);
	} else if (dynfreq_algo == DYNFREQ_ALGO_PID){
		return devfreq_update_algo_pid(dev);
	} else {
		return 0;
	}
}


#define gt_devfreq(DEV)	container_of((DEV), struct devfreq, dev)
static ssize_t store_level(struct device *dev, struct device_attribute *attr,
			  const char *buf, size_t count)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct power_userspace_data *data;
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;
	unsigned long wanted;
	unsigned long max_lvl = gtgov_data->userinfo.tbl_size - 1;
	int err = 0;

	mutex_lock(&devfreq->lock);
	data = &(gtgov_data->userinfo);

	sscanf(buf, "%lu", &wanted);
	if (wanted > max_lvl) {
		mutex_unlock(&devfreq->lock);
		err = count;
		goto err_range_out;
	}
	data->userlevel = wanted;
	data->valid = true;
	err = update_devfreq(devfreq);
	if (err == 0)
		err = count;
	mutex_unlock(&devfreq->lock);

	gtgov_notice("lvl update success, err = %d count=%lu cur userlevel[%lu]\n", err, count, data->userlevel);
	return err;

err_range_out:
	gtgov_notice("lvl update failed, for maxlevel is %ld, err = %d count=%lu cur userlevel[%lu]\n", max_lvl, err, count, data->userlevel);
	return err;
}

static ssize_t show_level(struct device *dev, struct device_attribute *attr,
			 char *buf)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct power_userspace_data *data;
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;
	int err = 0;

	mutex_lock(&devfreq->lock);
	data = &(gtgov_data->userinfo);

	if (data->valid)
		err = sprintf(buf, "%lu\n", data->userlevel);
	else
		err = sprintf(buf, "undefined\n");
	mutex_unlock(&devfreq->lock);
	return err;
}

static DEVICE_ATTR(level, 0644, show_level, store_level);

#define gt_devfreq(DEV)	container_of((DEV), struct devfreq, dev)
static ssize_t store_gt_devfreq_enable(struct device *dev, struct device_attribute *attr,
			  const char *buf, size_t count)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;

	if (!gtgov_data) {
		return count;
	}

	mutex_lock(&devfreq->lock);
	sscanf(buf, "%u", &gtgov_data->enable);
	mutex_unlock(&devfreq->lock);

	gtgov_notice("devfreq enable=%u\n", gtgov_data->enable);
	return count;
}

static ssize_t show_gt_devfreq_enable(struct device *dev, struct device_attribute *attr,
			 char *buf)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;
	int err = 0;

	if (!gtgov_data) {
		sprintf(buf, "undefined data\n");
		return 1;
	}

	mutex_lock(&devfreq->lock);
	err = sprintf(buf, "%u\n", gtgov_data->enable);
	mutex_unlock(&devfreq->lock);

	return err;
}
static DEVICE_ATTR(gt_devfreq_enable, 0644, show_gt_devfreq_enable, store_gt_devfreq_enable);

static ssize_t store_user_mode_powsave(struct device *dev, struct device_attribute *attr,
			  const char *buf, size_t count)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;
	unsigned int wanted;
	int err = 0;

	mutex_lock(&devfreq->lock);

	sscanf(buf, "%u", &wanted);
	gtgov_data->user_mode_powsave = wanted;
	err = update_devfreq(devfreq);
	if (err == 0)
		err = count;
	mutex_unlock(&devfreq->lock);

	gtgov_notice("err = %d count=%lu user_mode_powsave=%u\n", err, count, gtgov_data->user_mode_powsave);
	return err;
}

static ssize_t show_user_mode_powsave(struct device *dev, struct device_attribute *attr,
			 char *buf)
{
	struct devfreq *devfreq = gt_devfreq(dev);
	struct devfreq_gt_govdata *gtgov_data = devfreq->data;
	int err = 0;

	mutex_lock(&devfreq->lock);
	err = sprintf(buf, "%u\n", gtgov_data->user_mode_powsave);
	mutex_unlock(&devfreq->lock);

	return err;
}
static DEVICE_ATTR(user_mode_powsave, 0644, show_user_mode_powsave, store_user_mode_powsave);

static struct attribute *dev_entries[] = {
	&dev_attr_level.attr,
	&dev_attr_gt_devfreq_enable.attr,
	&dev_attr_user_mode_powsave.attr,
	NULL,
};

static const struct attribute_group dev_attr_group = {
	.name	= "user",
	.attrs	= dev_entries,
};

#if defined(__GT_DESKTOP__)
#define DEVICE_NAME "user_pwrctl"
struct user_pwrctl {
	struct miscdevice *osdev;
	struct devfreq *devfreq;
};
static struct user_pwrctl *user_pwrctl = NULL;
static int user_pwrctl_open(struct inode *inode, struct file *file)
{
	file->private_data = user_pwrctl;

	if (!user_pwrctl) {
		gtgov_err("failed: user_pwrctl is null\n");
		return -1;
	}

	return 0;
}

static int user_pwrctl_close(struct inode *inode, struct file *file)
{
	return 0;
}

static long user_pwrctl_set_pmmode(struct user_pwrctl *pwrctl, void __user *buf)
{
	unsigned int pmmod = 0;
	int err = 0;
	struct devfreq_gt_govdata *gov_data = NULL;

	if (gt_copy_from_user(&pmmod, buf, sizeof(pmmod))) {
		gtgov_err("copy from user failed!\n");
		return -EFAULT;
	}

	gov_data = pwrctl->devfreq->data;
	gov_data->user_mode_powsave = pmmod;
	mutex_lock(&pwrctl->devfreq->lock);
	err = update_devfreq(pwrctl->devfreq);
	if (err < 0) {
		gtgov_err("update_devfreq failed\n");
	}
	mutex_unlock(&pwrctl->devfreq->lock);
	gtgov_info("set pmmode is %u\n", pmmod);

	return 0;
}

static long user_pwrctl_get_pmmode(struct user_pwrctl *pwrctl, void __user *buf)
{
	unsigned int pmmode;
	struct devfreq_gt_govdata *gov_data = pwrctl->devfreq->data;

	pmmode = gov_data->user_mode_powsave;
	if (copy_to_user(buf, &pmmode, sizeof(unsigned int))) {
		gtgov_err("copy to user failed!\n");
		return -EFAULT;
	}

	return 0;
}

static int user_pwrctl_get_gpu_utils(struct user_pwrctl *pwrctl, void __user *buf)
{
	int util2user = 0;
	struct devfreq_gt_govdata *gov_data = pwrctl->devfreq->data;

	if (!gov_data->ppci_bdev) {
		gtgov_err("ppci_bdev is null\n");
		return -EFAULT;
	}

	util2user = hal_get_gpu_utils(gov_data->gt69srv_dev_node);
	if (copy_to_user(buf, &util2user, sizeof(int))) {
		gtgov_err("copy to user failed! util2user=%d\n", util2user);
		return -EFAULT;
	}
	gtgov_info("util2user = %d\n", util2user);
	return 0;
}

static int user_pwrctl_enable_stat_get(struct user_pwrctl *pwrctl, void __user *buf)
{
	struct devfreq_gt_govdata *gov_data = pwrctl->devfreq->data;
	unsigned enable = gov_data->enable;

	if (copy_to_user(buf, &enable, sizeof(unsigned int))) {
		gtgov_err("copy to user failed!=%u\n", enable);
		return -EFAULT;
	}
	gtgov_dbg("enable = %u\n", enable);

	return 0;
}

#define GT_PM_CMD 'p'
#define GT_PM_CMD_SET_PMMODE 		_IOW(GT_PM_CMD, 1, unsigned int)
#define GT_PM_CMD_GET_PMMODE 		_IOR(GT_PM_CMD, 2, unsigned int)
#define GT_PM_CMD_GET_GPU_UTILS 	_IOR(GT_PM_CMD, 3, int)
#define GT_PM_CMD_GET_ENABLE 		_IOR(GT_PM_CMD, 4, unsigned int)
static long user_pwrctl_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct user_pwrctl *pwrctl = (struct user_pwrctl *)file->private_data;

	switch (cmd) {
	case GT_PM_CMD_SET_PMMODE:
		return user_pwrctl_set_pmmode(pwrctl, (void __user *)arg);
	case GT_PM_CMD_GET_PMMODE:
		return user_pwrctl_get_pmmode(pwrctl, (void __user *)arg);
	case GT_PM_CMD_GET_GPU_UTILS:
		return user_pwrctl_get_gpu_utils(pwrctl, (void __user *)arg);
	case GT_PM_CMD_GET_ENABLE:
		return user_pwrctl_enable_stat_get(pwrctl, (void __user *)arg);
	default:
		gtgov_notice("cmd[%d] is not support\n", cmd);
		return -ENOTTY;

	}
}

static struct file_operations osdev_fops = {
	.owner = THIS_MODULE,
	.open = user_pwrctl_open,
	.release = user_pwrctl_close,
	.unlocked_ioctl = user_pwrctl_ioctl,
};

struct miscdevice osdev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = DEVICE_NAME,
	.fops = &osdev_fops
};

static int user_pwrctl_init(struct devfreq *devfreq)
{
	int ret;

	user_pwrctl = kmalloc(sizeof(struct user_pwrctl), GFP_KERNEL);
	if (gt_is_err_or_null(user_pwrctl)) {
		return -EINVAL;
	}

	user_pwrctl->devfreq = devfreq;
	user_pwrctl->osdev = &osdev;
	ret = misc_register(&osdev);
	if(ret < 0) {
		gtgov_err("misc register is error ret = %d\n", ret);
		goto osdev_register_failed;
	}

	gtgov_info("misc register is succeed\n");
	return 0;

osdev_register_failed:
	return ret;
}

static void user_pwrctl_exit(void)
{
	misc_deregister(&osdev);

	kfree(user_pwrctl);
}
#endif

static int userspace_init(struct devfreq *devfreq)
{
	int err = 0;

	err = sysfs_create_group(&devfreq->dev.kobj, &dev_attr_group);

	/*avoid server multi card register*/
#if defined(__GT_DESKTOP__)
	err = user_pwrctl_init(devfreq);
	if (err) {
		gtgov_err("user_pwrctl_init failed\n");
	}
#endif

	return err;
}

static void userspace_exit(struct devfreq *devfreq)
{
#if defined(__GT_DESKTOP__)
	user_pwrctl_exit();
#endif

	/*
	 * Remove the sysfs entry, unless this is being called after
	 * device_del(), which should have done this already via kobject_del().
	 */
	if (devfreq->dev.kobj.sd)
		sysfs_remove_group(&devfreq->dev.kobj, &dev_attr_group);
}

int devfreq_update_by_userelevel(struct devfreq * df)
{
	struct devfreq_gt_govdata *gtgov_data = df->data;
	struct power_userspace_data *data;
	unsigned long target_freq = 0;
	if (!gtgov_data) {
		gtgov_notice("gtgov_data is null\n");
		return 0;
	}

	data = &gtgov_data->userinfo;
	if (data->userlevel < data->tbl_size) {
		target_freq = data->level2freq[data->userlevel];
	} else {
		target_freq = gtgov_data->default_freq;
	}

	return target_freq;
}

static int devfreq_get_target_freq(struct devfreq *df,
					unsigned long *freq)
{
	unsigned long tfreq = 0;
	unsigned long user_setfreq = 0;
	unsigned long target_freq = 0;
	struct devfreq_gt_govdata *gtgov_data = df->data;

	if (!gtgov_data) {
		gtgov_notice("gtgov_data is null\n");
		return -1;
	}

	if (get_pwr_debug_lvl() >= PWRD_LVL_INFO) {
		gtgov_notice("devfreq %d is triggered\n",gtgov_data->id);
	}
	//get devfreq_private
	if (gtgov_data->curr_freq == 0) {
		gtgov_data->curr_freq = gtgov_data->default_freq;
	}

	/*enble pm & dvfs for debug*/
	if (!gtgov_data->enable) {
		*freq = gtgov_data->default_freq + gtgov_data->freq_offset;
		return 0;
	}

	//only dyn mode, update freq by temperature and gpu uselization
	if (gtgov_data->userinfo.user_pm_mode == HAL_GPU_WORK_MODE_DYN) {
		gt_idle_check(df, freq);
		if (gtgov_data->user_mode_powsave) {
			*freq = gtgov_data->user_powsave_freq + gtgov_data->freq_offset;
			gtgov_info("user_powsave_freq = %lu\n", *freq);
			return 0;
		}

		user_setfreq = devfreq_update_by_userelevel(df);
		tfreq = devfreq_update_by_thermal(df);

		if (user_setfreq == 0 && tfreq == 0) {
			target_freq = gtgov_data->curr_freq;
		} else {
			if (user_setfreq == 0) {
				target_freq = tfreq;
			} else if (tfreq == 0) {
				target_freq = user_setfreq;
			} else {
				target_freq = (user_setfreq < tfreq) ? user_setfreq : tfreq;
			}
		}
	}

	*freq = target_freq + gtgov_data->freq_offset;
	gtgov_data->curr_freq = *freq;
	//gtgov_err("%s: %d\n", __func__, __LINE__);

	return 0;
}

static void gt_devfreq_monitor_work(void *df)
{
	int err;
	struct devfreq* devfreq = (struct devfreq* )df;
	struct devfreq_gt_govdata *govdata = devfreq->data;

	mutex_lock(&devfreq->lock);
	err = update_devfreq(devfreq);
	if (err)
		gtgov_notice("dvfs failed with (%d) error\n", err);

	gt_queue_dwork(govdata->devfreq_wkq, govdata->devfreq_dwork,
				(devfreq->profile->polling_ms));
	mutex_unlock(&devfreq->lock);
}

static int devfreq_wkq = 0;
static void gt_devfreq_monitor_start(struct devfreq *devfreq)
{
	struct devfreq_gt_govdata *govdata = devfreq->data;

	/*init work*/
	sprintf(govdata->name, "%s-%d", "devfreq_wkq", devfreq_wkq);

	govdata->devfreq_wkq = gt_create_singlethread_wq("devfreq_wkq");
	if (IS_ERR_OR_NULL(govdata->devfreq_wkq)) {
		gtgov_notice("devfreq_wkq alloc is failed\n");
		goto err_end;
	}

	/*start temperature monitor*/
	govdata->devfreq_dwork = gt_dwork_alloc(gt_devfreq_monitor_work, devfreq);
	if (IS_ERR_OR_NULL(govdata->devfreq_dwork)) {
		gtgov_notice("failed to start temperature monitor\n");
		goto err_mwkq_alloc;
	}

	govdata->input_kick_dwork = gt_dwork_alloc(gt_input_kick_work, devfreq);
	if (IS_ERR_OR_NULL(govdata->input_kick_dwork)) {
		gtgov_notice("failed to start input event monitor\n");
		goto err_mdwk_alloc;
	}

	if (!gt_queue_dwork(govdata->devfreq_wkq, govdata->devfreq_dwork, 0)) {
		gtgov_notice("failed to start temperature monitor\n");
		goto err_inputwk_alloc;
	}

	gtgov_dbg("run finished start %s successfully\n", govdata->name);
	gt_devfreq_bind_input_handler(NULL, devfreq);
	devfreq_wkq++;

	return;

err_inputwk_alloc:
	gt_dwork_destroy(govdata->input_kick_dwork);
err_mdwk_alloc:
	gt_dwork_destroy(govdata->devfreq_dwork);
err_mwkq_alloc:
	gt_destroy_workqueue(govdata->devfreq_wkq);
err_end:
	govdata->input_kick_dwork = NULL;
	govdata->devfreq_dwork = NULL;
	govdata->devfreq_wkq = NULL;
	gtgov_notice("start %s failed\n", govdata->name);

	return;
}

static void gt_devfreq_monitor_stop(struct devfreq *devfreq)
{
	struct devfreq_gt_govdata *govdata = devfreq->data;

	gt_devfreq_unbind_input_handler();

	gt_cancel_dwork_sync(govdata->input_kick_dwork);

	gt_cancel_dwork_sync(govdata->devfreq_dwork);

	gt_flush_workqueue(govdata->devfreq_wkq);

	gt_dwork_destroy(govdata->input_kick_dwork);

	gt_dwork_destroy(govdata->devfreq_dwork);

	gt_destroy_workqueue(govdata->devfreq_wkq);

	gtgov_dbg("run finished\n");
}

struct gt_dwork {
	struct delayed_work dwork;
	void (*func)(void *data);
	void *data;
};

void gt_devfreq_monitor_resume(struct devfreq *devfreq)
{
	struct devfreq_gt_govdata *govdata = devfreq->data;
	unsigned long freq;
	gt_dwork *devfreq_dwork = govdata->devfreq_dwork;
	struct gt_dwork *dwk = (struct gt_dwork *)devfreq_dwork;
	struct delayed_work *dwork = &(dwk->dwork);

	mutex_lock(&devfreq->lock);
	if (!devfreq->stop_polling)
		goto out;

	if (!delayed_work_pending(dwork) &&
			devfreq->profile->polling_ms) {
		gt_queue_dwork(govdata->devfreq_wkq, govdata->devfreq_dwork, devfreq->profile->polling_ms);

		if (govdata->support_idle_switch) {
			gt_queue_dwork(govdata->devfreq_wkq, govdata->input_kick_dwork, 0);
			gt_input_resume();
		}
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0))
	devfreq->last_stat_updated = jiffies;
#else
	devfreq->stats.last_update = get_jiffies_64();
#endif
	devfreq->stop_polling = false;

	if (devfreq->profile->get_cur_freq &&
		!devfreq->profile->get_cur_freq(devfreq->dev.parent, &freq))
		devfreq->previous_freq = freq;

out:
	mutex_unlock(&devfreq->lock);
	gtgov_notice("resume done\n");
}
GT_EXT_SYM(gt_devfreq_monitor_resume);

void gt_devfreq_monitor_suspend(struct devfreq *devfreq)
{
	struct devfreq_gt_govdata *govdata = devfreq->data;

	mutex_lock(&devfreq->lock);
	if (devfreq->stop_polling) {
		mutex_unlock(&devfreq->lock);
		return;
	}

	devfreq_update_status(devfreq, devfreq->previous_freq);
	devfreq->stop_polling = true;
	mutex_unlock(&devfreq->lock);
	if (govdata->support_idle_switch) {
		gt_input_suspend();
		gt_cancel_dwork_sync(govdata->input_kick_dwork);
	}
	gt_cancel_dwork_sync(govdata->devfreq_dwork);
	gtgov_notice("suspend done\n");
}
GT_EXT_SYM(gt_devfreq_monitor_suspend);

static int devfreq_gt_handler(struct devfreq *devfreq,
				unsigned int event, void *data)
{
	switch (event) {
	case DEVFREQ_GOV_START:
		userspace_init(devfreq);
		gt_devfreq_monitor_start(devfreq);
		break;

	case DEVFREQ_GOV_STOP:
		userspace_exit(devfreq);
		gt_devfreq_monitor_stop(devfreq);
		break;
/*
	case DEVFREQ_GOV_INTERVAL:
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0))
		devfreq_interval_update(devfreq, (unsigned int *)data);
#else
		devfreq_update_interval(devfreq, (unsigned int *)data);
#endif
		break;
*/
	case DEVFREQ_GOV_SUSPEND:
		gt_devfreq_monitor_suspend(devfreq);
		break;

	case DEVFREQ_GOV_RESUME:
		gt_devfreq_monitor_resume(devfreq);
		break;

	default:
		break;
	}

	return 0;
}

struct devfreq_governor gt_devfreq_gov = {
	.name = "gt_dfreq_gov",
	.get_target_freq = devfreq_get_target_freq,
	.event_handler = devfreq_gt_handler,
};

int gt_devfreq_gov_register(void)
{
	return devfreq_add_governor(&gt_devfreq_gov);
}

int gt_devfreq_gov_unregister(void)
{
	int err = 0;

	err =  devfreq_remove_governor(&gt_devfreq_gov);
	gtgov_info("%s err = %d\n", err);

	return err;
}
#endif
