/*************************************************************************/ /*!
@File			gtdpu_dp.c
@Title
@Copyright		Copyright (c) Gengtu Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include <linux/platform_device.h>
#include <linux/compiler.h>
#include "gtdpu_dp.h"
#include "gtdpu_connector.h"
#include "gtdpu_dp_debugfs.h"
#ifdef CONFIG_DRM_GT_AUDIO
#include "gtaudio_chip_common.h"
#include "gtaudio_drv.h"
#endif

static int s_dp_monitor = 0x26b;
MODULE_PARM_DESC(s_dp_monitor,
		"\t\tBit 0 (0x01)  will enable edid segment pointer\n"
		"\t\tBit 1 (0x02)  will enable Listen for link status\n"
		"\t\tBit 2 (0x04)  will enable tramsmission configuration to improve signal quality\n"
		"\t\tBit 3 (0x08)  will enable ctx detect retry to ensure connector status\n"
		"\t\tBit 4 (0x10)  will enable lane count adjust\n"
		"\t\tBit 5 (0x20)  will enable mvid software control\n"
		"\t\tBit 6 (0x40)  will enable replacement of non-standard timing\n"
		"\t\tBIT 7 (0x80)  will enable display timing bit depth of the dp is obtained from the edid\n"
		"\t\tBIT 8 (0x100) will enable support for yuv output\n"
		"\t\tBIT 9 (0x200) will enable link power ctrl\n"
		"\t\tBIT 10(0x400) will enable alternate_scrambler_reset\n");
module_param_named(s_dp_monitor, s_dp_monitor, uint, 0600);

static int s_dp_swing_param = 0;
MODULE_PARM_DESC(s_dp_swing_param,
		"\t\tBit 0 - 7 Analogue parameter isel\n"
		"\t\tBit 8 -15 Analogue parameter mainsel\n"
		"\t\tBit 16-23 Analogue parameter post pre-emphasis\n"
		"\t\tBit 24-31 Analogue parameter pre  pre-emphisis\n");
module_param_named(s_dp_swing_param, s_dp_swing_param, uint, 0600);

static int dp_mdelay = 20;
MODULE_PARM_DESC(dp_mdelay, "msleep\n");
module_param_named(dp_mdelay, dp_mdelay, uint, 0600);

bool gt_dp_monitor_en(enum dp_monitor_status info)
{
	if (s_dp_monitor & GT_BIT(info))
		return true;

	return false;
}

static int gt_dp_hw_init(struct dp_device_t *gt_dp)
{
	if (gt_dp->chip.dp_hw_init)
		return gt_dp->chip.dp_hw_init(&gt_dp->chip);

	return 0;
}

static void gt_dp_hw_fini(struct dp_device_t *gt_dp)
{
	if (gt_dp->chip.dp_hw_fini)
		gt_dp->chip.dp_hw_fini(&gt_dp->chip);
}

static void gt_dp_clear_edid_info(struct dp_chip_t *chip)
{
	if (dp_get_edid(chip)) {
		dp_info(chip->dev, "clear edid\n");
		gt_kfree(dp_get_edid(chip));
		dp_set_edid(chip, NULL);
		gt_memset(chip->edid_buf, 0, sizeof(chip->edid_buf));
		chip->max_clk = 0;
	}
}

static void gt_dp_irq_handle(void *data)
{
	struct drm_crtc *crtc = NULL;
	int irq_status = gt_connector_status_unknown;
	struct dp_device_t *gt_dp = container_of(data, struct dp_device_t, chip);

	BUG_ON(!gt_dp);

	if (gt_dp->chip.dp_irq_handle)
		irq_status = gt_dp->chip.dp_irq_handle(data);


	if (gt_dp->connector.force == DRM_FORCE_ON ||
		gt_dp->connector.force == DRM_FORCE_ON_DIGITAL ||
		irq_status == gt_connector_status_connected) {
		dp_set_hpg_status(&gt_dp->chip, 1);
	} else if (irq_status == gt_connector_status_disconnected) {
		gt_dp_clear_edid_info(&gt_dp->chip);
		dp_set_hpg_status(&gt_dp->chip, 0);
	}

	if (irq_status != gt_connector_status_unknown) {
		crtc = gt_dp->encoder.crtc;
		dp_info(gt_dp->dev, "QueueHpdWork\n");
		if (!gt_queue_work(gt_dp->hpd_wq, gt_dp->hpd_work))
			dp_info(gt_dp->dev, "work item is pending execution!!!!!!\n");
	}
}

static void __attribute__((unused)) gt_dp_hw_irq_enable(struct dp_device_t *gt_dp)
{
	gtdpu_info(gt_dp->dev, DPU_UT_DP, "hw irq enable\n");

	if (gt_dp->chip.dp_irq_enable)
		gt_dp->chip.dp_irq_enable(&gt_dp->chip);
}

static void gt_dp_hw_irq_disable(struct dp_device_t *gt_dp)
{
	gtdpu_info(gt_dp->dev, DPU_UT_DP, "hw irq disable\n");

	if (gt_dp->chip.dp_irq_disable)
		gt_dp->chip.dp_irq_disable(&gt_dp->chip);
}

static int gt_dp_aux_rw(struct dp_device_t *gt_dp, struct aux_cfg *aux_rx)
{
	int drm_power_state = -1;
	struct dp_chip_t *chip = &gt_dp->chip;

	if (!aux_rx || !chip->dp_aux_channel_run)
		return DP_AUX_NATIVE_REPLY_NACK;

	if (gt_dp->panel) {
		if (gt_dp->drm_dev)
			drm_power_state = gt_dp->drm_dev->switch_power_state;

		if (drm_power_state == DRM_SWITCH_POWER_OFF) {
			gtdpu_info(chip->dev, DPU_UT_DP, "suspend...,disconnected\n");
			return DP_AUX_NATIVE_REPLY_NACK;
		}
	}

	gt_panel_prepare(chip->panel);

	if (dp_get_hpg_status(chip) == 0) {
		gtdpu_info(chip->dev, DPU_UT_DP, "disconnected\n");
		return DP_AUX_NATIVE_REPLY_NACK;
	}

	return chip->dp_aux_channel_run(aux_rx, chip);
}

static void dp_set_poll_cnt(struct dp_device_t *dp, int val)
{
	atomic64_set(&dp->poll_cnt, val);
}

static long long dp_get_poll_cnt(struct dp_device_t *dp)
{
	return atomic64_read(&dp->poll_cnt);
}

static const int gt_bw_table[] = {
	GTDP_LINK_BW_1_62,
	GTDP_LINK_BW_2_7,
	GTDP_LINK_BW_5_4,
	GTDP_LINK_BW_8_1,
};

static int gt_dp_rate_index(const int *rates, int len, int rate)
{
	int i;

	for (i = 0; i < len; i++)
		if (rate == rates[i])
			return i;

	return -1;
}

int gt_dp_mvid_calc(struct dp_chip_t *chip)
{
	int mvid = 0, mod = 0;
	struct dp_device_t *gt_dp = NULL;
	u64 temp1 = 0, f_ls_clk = 0, f_strm_clk = 0;

	if (!chip || !gt_dp_monitor_en(dp_mvid_sw_cacl))
		return -1;

	gt_dp = to_dp_device(chip);

	if (!gt_dp || gt_dp->current_mode.clock <= 0 || chip->lane_rate == 0)
		return -1;

	/* Link stream clock == pixel clock */
	f_strm_clk = gt_dp->current_mode.clock; /*KHZ*/

	/* Calculate Link Symbol clock */
	f_ls_clk = chip->lane_rate * 27000ULL; /*KHZ*/

	/* M / N == f_strm_clk / f_ls_clk
	 * M == f_strm_clk * N / f_ls_clk
	 */
	temp1 = f_strm_clk * ASYNC_CLOCK_MSA_NVID;
	mod = do_div(temp1, f_ls_clk);

	/* round up to five */
	mvid = temp1 + ((mod * 10ULL / f_ls_clk >= 5) ? 1 : 0);
	if (mvid >= (1<<24)) {
		mvid = -1;
		gtdpu_info(chip->dev, DPU_UT_DP, "[BAD] Invalid mvid\n");
	}

	gtdpu_info(chip->dev, DPU_UT_DP, "[GT DP] MSA-MVID:%d "
			"f_strm_clk:%lld KHZ f_ls_clk:%lld KHZ mod:%d",
			mvid, f_strm_clk, f_ls_clk, mod);

	return mvid;
}

int gt_dp_sink_power_ctrl(struct dp_device_t *gt_dp, bool power_on)
{
	u8 value = 0;
	int err = 0;

	/* DP_SET_POWER register is only available on DPCD v1.1 and later */
	if (!gt_dp || gt_dp->dpcd[DP_DPCD_REV] < DP_REV_11)
		return 0;

	if (!gt_dp_monitor_en(dp_link_pwr_ctl))
		return 0;

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "sink_power_ctrl:%s\n",
			power_on ? "power on" : "powern down");

	err = drm_dp_dpcd_read(&gt_dp->aux, DP_SET_POWER, &value, 1);
	if (err < 0)
		return err;

	value &= ~DP_SET_POWER_MASK;
	/* Sink - (State3 Sleep):
	 * 1. Hpd asserted
	 * 2. Aux enabled for differential signal monitoring,
	 * 3. Main-link Rx disabled
	 */

	/* Sink - (State 2 standby):
	 * 1. Hpd asserted
	 * 2. Aux enabled for differential signal monitoring,
	 * 3. Main-link Rx enabled
	 */

	value |= (power_on ? DP_SET_POWER_D0 : DP_SET_POWER_D3);

	err = drm_dp_dpcd_write(&gt_dp->aux, DP_SET_POWER, &value, 1);
	if (err < 0)
		return err;

	if (power_on) {
		/*
		 * According to the DP 1.1 specification, a "Sink Device must exit the
		 * power saving state within 1 ms" (Section 2.5.3.1, Table 5-52, "Sink
		 * Control Field" (register 0x600).
		 */
		if (gt_dp->is_R1) {
			gt_udelay(dp_mdelay * 1000);
		} else {
			if (gt_dp->use_rate_select) {
				 /* For an embedded connection, a Sink device may take up to 20 ms from a power-save mode
				  * until it is ready to reply to an AUX request transaction
				  */
				 gt_usleep_range(20000, 20100);
			} else {
				 gt_usleep_range(1000, 2000);
			}
		}
	} else {
		if (gt_dp->is_R1) {
			/* timing T10, at least 100ms */
			gt_udelay(100000);
		}
	}
	return 0;
}

static void gt_dp_rate_choose(struct dp_device_t *gt_dp)
{
	int i = 0;
	struct dp_chip_t *chip = &gt_dp->chip;

	/* For EDP 1.4 */
	if (gt_dp->use_rate_select) {
		chip->lane_rate = gt_dp->max_sink_rates;
		gtdpu_info(chip->dev, DPU_UT_DP, "use rate select:%#.8x\n", chip->lane_rate);
	}

	if (chip->lane_rate > chip->max_source_rate)
		chip->lane_rate = GTDP_LINK_BW_5_4;

	i = gt_dp_rate_index(gt_bw_table, GT_ARRAY_SIZE(gt_bw_table), chip->lane_rate);
	if (i < 0) {
		dp_info(chip->dev,
				"[BAD] Link rate matching failed using 5.4Gbps\n");
		chip->lane_rate = GTDP_LINK_BW_5_4;
		chip->phy_rate = 0x2;
	} else {
		chip->phy_rate = i;
	}
}

static void gt_dp_voltage_swing_adjust(struct dp_device_t *gt_dp)
{
	int j = 0;
	uint8_t train_set[4] = {0};

	if (!gt_dp)
		return;

	for (j = 0; j < gt_dp->chip.lane_count; j++) {
		train_set[j] = gt_dp->chip.lane_swing[j];
		if (gt_dp->chip.lane_swing[j] >= DP_TRAIN_VOLTAGE_SWING_LEVEL_2)
			train_set[j] |= DP_TRAIN_MAX_SWING_REACHED;

		train_set[j] |= (gt_dp->chip.lane_emphasis[j]);
		if (gt_dp->chip.lane_emphasis[j] >= DP_TRAIN_PRE_EMPH_LEVEL_2)
			train_set[j] |= DP_TRAIN_MAX_PRE_EMPHASIS_REACHED;
	}

	if (gt_dp_monitor_en(dp_swing_adjust) && gt_dp->chip.dp_source_swing_adjust) {
		gt_dp->chip.dp_source_swing_adjust(&gt_dp->chip, s_dp_swing_param);
	}

	drm_dp_dpcd_write(&gt_dp->aux, DP_TRAINING_LANE0_SET,
		train_set, gt_dp->chip.lane_count);
}

static void gt_dp_get_adjust_train(struct dp_device_t *gt_dp,
		const uint8_t link_status[DP_LINK_STATUS_SIZE])
{
	int lane = 0;

	if (!gt_dp)
		return;

	for (lane = 0; lane < gt_dp->chip.lane_count; lane++) {
		gt_dp->chip.lane_swing[lane] = drm_dp_get_adjust_request_voltage(link_status, lane);
		gt_dp->chip.lane_emphasis[lane] = drm_dp_get_adjust_request_pre_emphasis(link_status, lane);
		gtdpu_info(gt_dp->dev, DPU_UT_DP,
				"adjust request lane%d:lane_swing:%d lane_emphasis:%d ",
				lane, gt_dp->chip.lane_swing[lane],
				gt_dp->chip.lane_emphasis[lane] >> DP_TRAIN_PRE_EMPHASIS_SHIFT);
	}
}

static void gt_edp_caps_set(struct dp_device_t *gt_dp)
{
	u8 rate_select = 0, scramble_reset = 0;
	int rate_index = 0;

	if (!gt_dp)
		return;

	if (gt_dp->use_rate_select) {

		gt_dp_voltage_swing_adjust(gt_dp);

		rate_index = gt_dp_rate_index(gt_dp->sink_rates,
					gt_dp->num_sink_rates, gt_dp->chip.lane_rate);
		if (gt_warn_on(rate_index < 0))
			rate_index = 0;

		rate_select = rate_index;

		drm_dp_dpcd_write(&gt_dp->aux, DP_LINK_RATE_SET,
				&rate_select, 1);

		gtdpu_info(gt_dp->dev, DPU_UT_DP, "EDP Rate Set[%#x]\n", gt_dp->chip.lane_rate);

		if (gt_dp->chip.alternate_scrambler_reset) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "dp alternate scrambler reest enable\n");
			scramble_reset = DP_ALTERNATE_SCRAMBLER_RESET_ENABLE;
			drm_dp_dpcd_write(&gt_dp->aux, DP_EDP_CONFIGURATION_SET, &scramble_reset, 1);
		}
	}
}

static void gt_edp_init_dpcd(struct dp_device_t *gt_dp)
{
	int i = 0;
	int val = 0;
	__le16 sink_rates[DP_MAX_SUPPORTED_RATES];

	if (!gt_dp)
		return;

	/*
	 * Read the eDP display control registers.
	 *
	 * Do this independent of DP_DPCD_DISPLAY_CONTROL_CAPABLE bit in
	 * DP_EDP_CONFIGURATION_CAP, because some buggy displays do not have it
	 * set, but require eDP 1.4+ detection (e.g. for supported link rates
	 * method). The display control registers should read zero if they're
	 * not supported anyway.
	 */
	if (drm_dp_dpcd_read(&gt_dp->aux, DP_EDP_DPCD_REV,
			     gt_dp->edp_dpcd, sizeof(gt_dp->edp_dpcd)) ==
			     sizeof(gt_dp->edp_dpcd)) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP,"eDP DPCD: %*ph\n",
			    (int)sizeof(gt_dp->edp_dpcd),
			    gt_dp->edp_dpcd);
	}

	/* Clear the default sink rates */
	gt_dp->num_sink_rates = 0;
	gt_dp->max_sink_rates = 0;

	/* Read the eDP 1.4+ supported link rates. */
	if (gt_dp->edp_dpcd[0] >= DP_EDP_14) {
		drm_dp_dpcd_read(&gt_dp->aux, DP_SUPPORTED_LINK_RATES,
				sink_rates, sizeof(sink_rates));

		for (i = 0; i < GT_ARRAY_SIZE(sink_rates); i++) {
			val = le16_to_cpu(sink_rates[i]);

			if (val == 0)
				break;

			/* Value read multiplied by 200kHz gives the per-lane
			 * link rate in kHz.
			 */
			gt_dp->sink_rates[i] = (val * 200) / 270000;
			if (gt_dp->max_sink_rates < gt_dp->sink_rates[i] &&
				gt_dp->chip.max_source_rate >= gt_dp->sink_rates[i] &&
				gt_dp_rate_index(gt_bw_table, GT_ARRAY_SIZE(gt_bw_table), gt_dp->sink_rates[i]) >= 0) {
				gt_dp->max_sink_rates = gt_dp->sink_rates[i];
			}

			gtdpu_info(gt_dp->dev, DPU_UT_DP, "edp sink rates-%#.8x\n", gt_dp->sink_rates[i]);
		}
		gt_dp->num_sink_rates = i;
	}

	/*
	 * Use DP_LINK_RATE_SET if DP_SUPPORTED_LINK_RATES are available,
	 * default to DP_MAX_LINK_RATE and DP_LINK_BW_SET otherwise.
	 */
	if (gt_dp->num_sink_rates) {
		gt_dp->use_rate_select = true;
		gt_dp->connector.connector_type = DRM_MODE_CONNECTOR_eDP;
		gt_dp->connector.force = DRM_FORCE_ON;
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "edp max sink rate-%#.8x\n", gt_dp->max_sink_rates);
	} else {
		gt_dp->use_rate_select = false;
	}
}

static bool gt_dp_rate_valid(int lane_rate, int lane_count, int clock, int bpc)
{
	unsigned long requirement, capacity;

	capacity = lane_rate * 27 * 1000 * 8 * lane_count;
	requirement = clock * bpc * 3;

	if (capacity >= requirement)
		return true;

	return false;
}

static const unsigned int gt_misc0[][5] = {
	{
	/* Legacy RGB mode*/
		0x0,  /* 6bpc */
		0x20, /* 8bpc */
		0x40, /* 10bpc*/
		0x60, /* 12bpc*/
		0x80, /* 16bpc*/
//	}, {
//	/* CEA RGB */
//		0x8,  /* 6bpc */
//		0x28, /* 8bpc */
//		0x48, /* 10bpc*/
//		0x68, /* 12bpc*/
//		0x88, /* 16bpc*/
	}, {
	/* Ycbcr 444 */
		0,	  /* unsupport 6bpc */
		0x2c, /* 8bpc */
		0x4c, /* 10bpc*/
		0x6c, /* 12bpc*/
		0x8c, /* 16bpc*/
	}, {
	/* Ycbcr 422 */
		0,	  /* unsupport 6bpc */
		0x2a, /* 8bpc */
		0x4a, /* 10bpc*/
		0x6a, /* 12bpc*/
		0x8a, /* 16bpc*/
	},
};

static const unsigned int gt_dp_bpc[5] = {6, 8, 10, 12, 16};

static int gt_dp_bpc_choose(struct dp_device_t *gt_dp)
{
	int index = 0;
	struct dp_display_info *info = &gt_dp->chip.display_info;

	index = gt_dp_rate_index(gt_dp_bpc, GT_ARRAY_SIZE(gt_dp_bpc), info->bpc);
	if (index < 0) {
		index = 1;
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "[BAD]Invalid bpc, Using 8 bpc\n");
	}

	while (index >= 0) {
		if (gt_dp_rate_valid(gt_dp->chip.lane_rate, gt_dp->chip.lane_count,
			gt_dp->current_mode.clock, gt_dp_bpc[index]))
			break;

		gtdpu_info(gt_dp->dev, DPU_UT_DP, "[BAD]Invalid bpp, Try to lower the bit depth\n");
		index--;
	}

	if (index < 0) {
		gtdpu_err(gt_dp->dev, "Timing of dp exceeds maximum limit\n");
		index = 1;
	}

	return index;
}

static void gt_dp_video_timing_prepare(struct dp_device_t *gt_dp)
{
	int index = 0;
	struct dp_display_info *info = &gt_dp->chip.display_info;
	const struct drm_display_info *drm_info = &gt_dp->connector.display_info;

	/* default:
	 * bpc = 8
	 * color formats: RGB4:4:4
	*/
	info->bpc = 8;
	info->color_formats = GT_COLOR_FORMAT_RGB444;

	if (gt_dp_monitor_en(dp_bpc_use_edid)) {
		info->bpc = drm_info->bpc;
		info->color_formats = GT_COLOR_FORMAT_RGB444;
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "Maximum bpc supported by the monitor:%d\n", info->bpc);

		if (gt_dp_monitor_en(dp_yuv_support)) {
			if (drm_info->color_formats & GT_COLOR_FORMAT_YCBCR444) {
				info->color_formats = GT_COLOR_FORMAT_YCBCR444;
				gtdpu_info(gt_dp->dev, DPU_UT_DP, "color formats:YUV444\n");
			} else if (drm_info->color_formats & GT_COLOR_FORMAT_YCBCR422) {
				info->color_formats = GT_COLOR_FORMAT_YCBCR422;
				gtdpu_info(gt_dp->dev, DPU_UT_DP, "color formats:YUV422\n");
			}
		}
	}

	index = gt_dp_bpc_choose(gt_dp);
	info->bpc = gt_dp_bpc[index];
	info->misc0 = gt_misc0[info->color_formats >> 1][index];
	info->video_map = index + GT_COLOR_FORMAT_YCBCR422 * (info->color_formats >> 1);

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "gt_dp display info"
			"[bpc:%d misc0:%#x video_map:%d]\n",
			info->bpc, info->misc0, info->video_map);
}

static void gt_dp_lane_choose(struct dp_device_t *gt_dp)
{
	const int lanes[3] = {1, 2, 4};
	char buf[MONITOR_NAME_LEN] = {0};
	struct dp_chip_t *chip = &gt_dp->chip;
	/* Unis displays flicker at 2560x1400@75HZ and 1600x900@60HZ,
	 * lowering the number of links can fix it!
	 * */
	const char *monitor_array[] = {
		"Unis V271Q"
	};

	gt_dp->reduce_lane_count = false;

	/* Maximum number of links used by default */
	if (gt_dp_rate_index(lanes, GT_ARRAY_SIZE(lanes), chip->lane_count) < 0) {
		chip->lane_count = 0x4;
		gt_dp->reduce_lane_count = true;
		dp_info(gt_dp->dev,
				"[BAD] Link count matching failed using 4 lane\n");
	}

	if (dp_get_edid(chip) && !gt_dp_monitor_en(dp_lanes_adjust)) {
		drm_edid_get_monitor_name(dp_get_edid(chip), buf, GT_ARRAY_SIZE(buf));
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "monitor-%s\n", buf);

		if (match_string(monitor_array, GT_ARRAY_SIZE(monitor_array), buf) >= 0) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "monitor-reduce lane_count!\n");
			gt_dp->reduce_lane_count = true;
		}
	}

	if ((gt_dp->reduce_lane_count || gt_dp_monitor_en(dp_lanes_adjust)) && (chip->lane_count > 1)) {
		if (gt_dp_rate_valid(chip->lane_rate, chip->lane_count >> 1,
			gt_dp->current_mode.clock, gt_dp_bpc[1])) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "reduce the number of lanes\n");
			chip->lane_count = chip->lane_count >> 1;
		}
	}

	gt_dp->chip.phy_lanes = gt_dp_rate_index(lanes, GT_ARRAY_SIZE(lanes), chip->lane_count);
}

static void gt_dp_rate_adjust(struct dp_device_t *gt_dp, bool adjust)
{
	int i = 0;
	u8 reduce_phy_rate = 0;
	u8 reduce_lane_rate = 0;

	if (adjust && gt_dp->current_mode.clock) {
		i = gt_dp_rate_index(gt_bw_table, GT_ARRAY_SIZE(gt_bw_table), gt_dp->chip.lane_rate);
		if (i > 0 && i < GT_ARRAY_SIZE(gt_bw_table)) {
			reduce_lane_rate = gt_bw_table[i - 1];
			reduce_phy_rate = i - 1;
		} else {
			return;
		}

		if (gt_dp->use_rate_select)
			i = gt_dp_rate_index(gt_dp->sink_rates,
					gt_dp->num_sink_rates, reduce_lane_rate);

		if (gt_dp_rate_valid(reduce_lane_rate, gt_dp->chip.lane_count,
			gt_dp->current_mode.clock, gt_dp->chip.display_info.bpc) && i >= 0) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "reduce bit rate!!!!!!\n");
			gtdpu_info(gt_dp->dev, DPU_UT_DP,
					"old:lane_rate-%#x, phy_rate-%#d"
					"new:lane_rate-%#x, phy_rate-%d\n",
					gt_dp->chip.lane_rate,gt_dp->chip.phy_rate,
					reduce_lane_rate, reduce_phy_rate);

			gt_dp->chip.lane_rate = reduce_lane_rate;
			gt_dp->chip.phy_rate = reduce_phy_rate;
		}
	}
}

static int gt_dp_caps_prepare(struct dp_device_t *gt_dp)
{
	if (!gt_dp)
		return -EFAULT;

	/* sink - max link rate */
	gt_dp->chip.lane_rate = gt_dp->dpcd[DP_MAX_LINK_RATE];
	/* sink - max link count */
	gt_dp->chip.lane_count = gt_dp->dpcd[DP_MAX_LANE_COUNT] & DP_MAX_LANE_COUNT_MASK;

	/* sink - enhanced mode support */
	gt_dp->chip.enhance_mode = drm_dp_enhanced_frame_cap(gt_dp->dpcd) ? 1 : 0;

	gt_dp->chip.tps3_support = drm_dp_tps3_supported(gt_dp->dpcd);

	/* sink - ASSR support, For edp devices only */
	if (gt_dp_monitor_en(alternate_scrambler_reset)) {
		if (gt_dp->use_rate_select) {
			gt_dp->chip.alternate_scrambler_reset = gt_dp->dpcd[DP_EDP_CONFIGURATION_CAP] & DP_ALTERNATE_SCRAMBLER_RESET_CAP;
		} else {
			gt_dp->chip.alternate_scrambler_reset = 0;
		}
	} else {
		gt_dp->chip.alternate_scrambler_reset = 0;
	}
	gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"DpcdRev-%#x MaxLineCount-%d MaxLinkRate-%#x ASSR-%d Enhance-%d Tps3-%s\n",
			gt_dp->dpcd[DP_DPCD_REV],
			gt_dp->chip.lane_count,
			gt_dp->chip.lane_rate,
			gt_dp->chip.alternate_scrambler_reset,
			gt_dp->chip.enhance_mode,
			gt_dp->chip.tps3_support ? "ture" : "false");

	/* Match to a link rate supported by source */
	gt_dp_rate_choose(gt_dp);

	/* Adjust the numger of links in the source to avoid flickering of
	 * individual monitor with specific resolutions
	 * */
	gt_dp_lane_choose(gt_dp);

	gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"using %u lanes at %u kHz (phy lanes:%d phy rates:%d)\n",
			gt_dp->chip.lane_count, gt_dp->chip.lane_rate * 27000,
			gt_dp->chip.phy_lanes, gt_dp->chip.phy_rate);

	return 0;
}

void gt_dp_check_sink_connection(struct dp_device_t *gt_dp)
{
	unsigned long timeout = 0;
	int ret = 0, retry = 0, status = 0;

	if (!gt_dp->chip.connector_detect) {
		gtdpu_err(gt_dp->chip.dev, "ConnectorDetect NULL!!!");
		return;
	}

	if ((gt_dp->connector.force == DRM_FORCE_ON ||
		 gt_dp->connector.force == DRM_FORCE_ON_DIGITAL) &&
		gt_dp->connected) {
		dp_set_hpg_status(&gt_dp->chip, 1);
		return;
	}

	dp_info(gt_dp->dev, "sink connection check\n");

	gt_mutex_lock(gt_dp->aux_mutex);

	/*Minimum TImes after asserting HPD at the end of IRQ_HPD:2ms */
	gt_usleep_range(5000, 6000);

	/*
	 * Attempt to read sink revision, retry in case the sink may not be ready.
	 *
	 * Sinks are *supposed* to come up within 1ms from an off state, but
	 * some docks need more time to power up.
	 */

	status = gt_dp->chip.connector_detect(&gt_dp->chip);
	if (status == gt_connector_status_connected)
		gt_usleep_range(15000, 20000);

	if (gt_dp->connector.force == DRM_FORCE_ON ||
		 gt_dp->connector.force == DRM_FORCE_ON_DIGITAL) {
		dp_set_hpg_status(&gt_dp->chip, 1);
	}

	timeout = jiffies + gt_msecs_to_jiffies(2000);

	while (gt_time_before(timeout)) {

		if (dp_get_hpg_status(&gt_dp->chip) == 0) {
			dp_info(gt_dp->dev, "HpdDisconnect\n");
			gt_dp->connected = false;
			break;
		}

		ret = drm_dp_dpcd_read(&gt_dp->aux, DP_DPCD_REV, gt_dp->dpcd, DP_RECEIVER_CAP_SIZE);
		if (ret == DP_RECEIVER_CAP_SIZE &&
			gt_dp->dpcd[DP_DPCD_REV] >= DP_REV_10 &&
			gt_dp->dpcd[DP_DPCD_REV] <= DP_REV_14) {
			gt_dp->connected = true;

			gt_edp_init_dpcd(gt_dp);

			break;
		} else {
			dp_info(gt_dp->dev,
					"sink connection check-retry[%d]\n", ++retry);
			gt_dp->connected = false;
		}

		gt_usleep_range(5000, 10000);
	}

	dp_info(gt_dp->dev,
			"sink connection check end[%s]\n", gt_dp->connected ? "connect" : "disconnect");

	gt_mutex_unlock(gt_dp->aux_mutex);
}

static int gt_dp_detect_ctx(struct drm_connector *connector,
		struct drm_modeset_acquire_ctx *ctx, bool force)
{
	struct dp_device_t *gt_dp = to_dp_device(connector);
	int status = gt_connector_status_unknown;

	if (!gt_dp)
		return -EFAULT;

	if (!gt_dp->chip.connector_detect) {
		gtdpu_err(gt_dp->dev, "Not support %s chip detect handle !!!\n",
				gt_dp->name);
		return -EFAULT;
	}

	if (connector->force == DRM_FORCE_ON ||
		connector->force == DRM_FORCE_ON_DIGITAL) {
		return connector_status_connected;
	}

	gt_mutex_lock(gt_dp->aux_mutex);

	status = gt_dp->chip.connector_detect(&gt_dp->chip);
	if (status == gt_connector_status_connected) {
		if (gt_dp_monitor_en(dp_rxctx_detect) && !gt_dp->connected) {
			dp_info(gt_dp->dev,
					"[BAD]hpd is detected, but aux channel no answer.\n");
			status = gt_connector_status_disconnected;
			/**
			 * hpd pull high, aux read and write normal, do not need to reset ctrl trigger hpd signal,
			 * only hpd pull high, aux can not read and write, the software triggers once the source
			 * side of the hpd signal to circumvent the:
			 * 1. after the first power on, the source side of the aux initialisation abnormalities
			 * 	caused by the connection state detection error
			 * 2. individual monitors can still detect the hpd signal after switching off the monitor
			 * 	power supply. but aux can not read and write
			 */
			if (gt_dp->chip.dp_hpd_signal_retrigger && !gt_dp->hpd_notify &&
				READ_ONCE(gt_dp->connector.dpms) != DRM_MODE_DPMS_ON) {
				dp_info(gt_dp->dev,
						"hpd signal retrigger.\n");
				gt_dp->hpd_notify = true;
				/**
				 * Note-
				 * dp hpd signal retrigger will reset ctrl, video timing is restored to the default
				 * value, does not affect phy, aux read and write
				 */
				gt_dp->chip.dp_hpd_signal_retrigger(&gt_dp->chip);
			}
			goto end_detect;
		}
	}

	gt_dp->hpd_notify = false;

end_detect:
	gt_mutex_unlock(gt_dp->aux_mutex);
	return status;
}

static enum drm_connector_status gt_dp_detect(struct drm_connector *connector, bool force)
{
    return gt_dp_detect_ctx(connector, NULL, force);
}

static bool gt_dp_clock_recover(struct dp_device_t *gt_dp)
{
	int i = 0;
	int cdr_delay = 0;
	uint8_t link_config[2];
	uint8_t link_status[DP_LINK_STATUS_SIZE];

	if (gt_dp->chip.dp_training_pattern_set)
		gt_dp->chip.dp_training_pattern_set(&gt_dp->chip, DP_TRAINING_PATTERN_1);

	gt_edp_caps_set(gt_dp);

	/* Write the link configuration data */

	link_config[0] = gt_dp->chip.lane_rate; /* DP_LINK_BW_SET */
	link_config[1] = gt_dp->chip.lane_count | (gt_dp->chip.enhance_mode << 7); /* DP_LANE_COUNT_SET */
	drm_dp_dpcd_write(&gt_dp->aux, DP_LINK_BW_SET, link_config, 2);

	link_config[0] = 0;
	link_config[1] = DP_SET_ANSI_8B10B;
	drm_dp_dpcd_write(&gt_dp->aux, DP_DOWNSPREAD_CTRL, link_config, 2);

	link_config[0] = DP_TRAINING_PATTERN_1; /* DP_TRAINING_PATTERN_SET */
	drm_dp_dpcd_write(&gt_dp->aux, DP_TRAINING_PATTERN_SET, link_config, 1);

	cdr_delay = gt_dp->dpcd[DP_TRAINING_AUX_RD_INTERVAL] & 0x7f;
	if (cdr_delay <= 0) {
		cdr_delay = 4000;
	} else {
		cdr_delay *= 4000; /* us */
	}
	gtdpu_info(gt_dp->dev, DPU_UT_DP, "clock recover delay %d us\n", cdr_delay);

	for (i = 0; i < 5; i++) {

		gt_usleep_range(cdr_delay, cdr_delay + 100);

		if (drm_dp_dpcd_read(&gt_dp->aux, DP_LANE0_1_STATUS, link_status,
			DP_LINK_STATUS_SIZE) != DP_LINK_STATUS_SIZE) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "[BAD]failed to get link status\n");
			return false;
		}

		gtdpu_info(gt_dp->dev, DPU_UT_DP, "lane0-1 status:%#.2x lane2-3 status:%#.2x\n",
				link_status[0], link_status[1]);

		if (drm_dp_clock_recovery_ok(link_status, gt_dp->chip.lane_count)) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "CDR Training Succeeded at %d Loop.\n", i + 1);
			break;
		} else {
			gt_dp_get_adjust_train(gt_dp, link_status);
			gt_dp_voltage_swing_adjust(gt_dp);
		}
	}

	return true;
}

static bool gt_dp_channel_balance(struct dp_device_t *gt_dp)
{
	int i = 0;
	int eq_delay = 0;
	uint8_t link_config[2];
	int pattern = DP_TRAINING_PATTERN_2;
	uint8_t link_status[DP_LINK_STATUS_SIZE];

	if (gt_dp->chip.tps3_support)
		pattern = DP_TRAINING_PATTERN_3;

	if (gt_dp->chip.dp_training_pattern_set)
		gt_dp->chip.dp_training_pattern_set(&gt_dp->chip, pattern);

	link_config[0] = pattern; /* DP_TRAINING_PATTERN_SET */
	drm_dp_dpcd_write(&gt_dp->aux, DP_TRAINING_PATTERN_SET, link_config, 1);

	eq_delay = gt_dp->dpcd[DP_TRAINING_AUX_RD_INTERVAL] & 0x7f;
	if (eq_delay <= 0) {
		eq_delay = 4000;
	} else {
		eq_delay *= 4000; /* us */
	}

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "eq delay %d us\n", eq_delay);

	for (i = 0; i < 5; i++) {

		gt_usleep_range(eq_delay, eq_delay + 100);

		if (drm_dp_dpcd_read(&gt_dp->aux, DP_LANE0_1_STATUS, link_status,
					DP_LINK_STATUS_SIZE) != DP_LINK_STATUS_SIZE) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "[BAD]failed to get link status\n");
			return false;
		}

		gtdpu_info(gt_dp->dev, DPU_UT_DP, "lane0-1 status:%#.2x lane2-3 status:%#.2x\n",
				link_status[0], link_status[1]);

		if (drm_dp_channel_eq_ok(link_status, gt_dp->chip.lane_count)) {
			gtdpu_info(gt_dp->dev, DPU_UT_DP, "EQ Training Succeeded at %d Loop.\n", i + 1);
			break;
		} else {
			gt_dp_get_adjust_train(gt_dp, link_status);
			gt_dp_voltage_swing_adjust(gt_dp);
		}
	}

	return true;
}

static void gt_dp_link_start(struct dp_device_t *gt_dp)
{
	int i = 0;
	uint8_t link_config[9] = {0};

	if (gt_dp->chip.dp_training_pattern_set)
		gt_dp->chip.dp_training_pattern_set(&gt_dp->chip, DP_TRAINING_PATTERN_DISABLE);

	link_config[0] = DP_TRAINING_PATTERN_DISABLE;
	drm_dp_dpcd_write(&gt_dp->aux, DP_TRAINING_PATTERN_SET, link_config, 1);

	gt_usleep_range(10000, 11000);

	drm_dp_dpcd_read(&gt_dp->aux, DP_LINK_BW_SET, link_config, sizeof(link_config));

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "Set Rx Lane rate :%#.2x\n", link_config[0]);
	gtdpu_info(gt_dp->dev, DPU_UT_DP, "Set Rx Lane count:%#.2x\n", link_config[1] & DP_LANE_COUNT_MASK);
	gtdpu_info(gt_dp->dev, DPU_UT_DP, "Set Rx enhanced_frame_en:%#.1x\n",
			(link_config[1] & DP_LANE_COUNT_ENHANCED_FRAME_EN));

	for (i = 0; i < 4; i++) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"Training lane-%d set "
			"voltage_swing:%#.2x max_swing_reached:%d "
			"lane_emphasis :%#.2x max_emphasis_reached:%d\n", i,
			(link_config[i + 3]) & 0x3,
			(link_config[i + 3] >> 2) & 0x1,
			(link_config[i + 3] >> 3) & 0x3,
			(link_config[i + 3] >> 5) & 0x1);
	}

	drm_dp_dpcd_read(&gt_dp->aux, DP_LANE0_1_STATUS, link_config, sizeof(link_config));

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "lane0-1 status:%#.2x lane2-3 status:%#.2x\n",
			link_config[0], link_config[1]);
}

int sink_lane_status_get(struct dp_device_t *gt_dp, u8 status[DP_LINK_STATUS_SIZE])
{
	int ret = 0;

	if (drm_dp_dpcd_read_link_status(&gt_dp->aux, status) == DP_LINK_STATUS_SIZE) {
		ret = DP_AUX_NATIVE_REPLY_ACK;
	} else {
		ret = DP_AUX_NATIVE_REPLY_NACK;
	}

	return ret;
}

static bool gt_dp_need_link_train(struct dp_device_t *gt_dp)
{
	bool ret = false;
	u8 link_status[DP_LINK_STATUS_SIZE];

	if (dp_get_hpg_status(&gt_dp->chip) == 0) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "dp disconnected\n");
		return ret;
	}

	if (gt_dp->encoder.crtc &&
		gt_dp->encoder.crtc->state &&
		gt_dp->encoder.crtc->state->active) {
		if (sink_lane_status_get(gt_dp, link_status) == 0) {
			if (!drm_dp_clock_recovery_ok(link_status, gt_dp->chip.lane_count) ||
				!drm_dp_channel_eq_ok(link_status, gt_dp->chip.lane_count) ||
				!gt_dp_rate_valid(gt_dp->chip.lane_rate, gt_dp->chip.lane_count,
					gt_dp->current_mode.clock, gt_dp->chip.display_info.bpc)) {
				ret = true;
			}
		}
	} else {
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "crtc disable\n");
	}

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "lane_count-%d lane_status-%*ph"
		" mode.clock-%d KHZ link_train-%s\n",
		gt_dp->chip.lane_count, sizeof(link_status), link_status,
		gt_dp->current_mode.clock, ret ? "true" : "false");

	return ret;
}

static void gt_dp_update_audio(struct dp_device_t *gt_dp)
{
	if (!gt_dp || !gt_dp->ac)
		return;

#ifdef CONFIG_DRM_GT_AUDIO
	if(gt_dp->ac->report_jack) {
		int conn_st, dpms;
		dpms = gt_dp->connector.dpms;
		conn_st = gt_dp->connected;
		gt_dp->ac->has_audio = dp_get_audio_status(&gt_dp->chip);
		gt_dp->ac->update_eld(gt_dp->ac, gt_dp->connector.eld, MAX_ELD_BYTES);
		gt_dp->ac->report_jack(gt_dp->ac, dpms, conn_st);
	}
#endif
}

bool gt_dp_is_skip_replace(const struct drm_display_mode *mode)
{
	struct resolution_info resolution[] = {
		RESOLUTION_INFO_ITEM(3840, 2160, -1, 533250),
		RESOLUTION_INFO_ITEM(3840, 2160, -1, 537600),
		RESOLUTION_INFO_ITEM(3840, 2160, -1, 573410),
		RESOLUTION_INFO_ITEM(3840, 2160, -1, 529910),
	};
	struct resolution_info  *match_resolution = NULL;

	if (!mode)
		return false;

	match_resolution = gtdpu_resolution_match((const gt_drm_display_mode *)mode, resolution, GT_ARRAY_SIZE(resolution));
	if (match_resolution) {
		if (s_dpu_debug & DPU_UT_DP) {
			gt_printk(KERN_INFO"match resolution(%dx%d@%d %dKHz), skip replace\n",
				match_resolution->hdisplay, match_resolution->vdisplay,
				match_resolution->vrefresh, match_resolution->clock);
		}
		return true;
	}

	return false;
}

static void gt_dp_hpd_work(void *data)
{
	struct dp_device_t *gt_dp = (struct dp_device_t *)data;
	int dpms = READ_ONCE(gt_dp->connector.dpms);
	struct dp_chip_t *chip = &gt_dp->chip;
	int drm_power_state = -1;
	bool changed = false;

	if (gt_dp->drm_dev)
		drm_power_state = gt_dp->drm_dev->switch_power_state;

	gtdpu_info(chip->dev, DPU_UT_DP,
			"dp hpd work [drm_power_state=%d][dpms=%d][old_connector:%d][crtc:%s]\n",
			drm_power_state, dpms, gt_dp->connector.status,
			gt_dp->encoder.crtc ? "true" : "fales");

	/* 1. If the hot-plugging is triggered during the wake-up process of s4 and s3,
	 * at this time, the resume of drm has not been called, then ignore the report uevent.
	 *
	 * 2. If the display is HDMI only or VGA only(DP is turned off, but DP is connected),
	 * uevent reporting is ignored(Generate insertion interrupt within 1s).
	 * */

	cancel_delayed_work_sync(&gt_dp->irq_event_work);
	dp_set_poll_cnt(gt_dp, 0);
	cancel_delayed_work_sync(&gt_dp->dp_poll_work);

	gt_dp_check_sink_connection(gt_dp);

	if (drm_power_state == DRM_SWITCH_POWER_ON) {
		if (gt_dp->connector.status == gt_connector_status_connected &&
			dpms == 3 && dp_get_hpg_status(chip) == 0 && !gt_dp->encoder.crtc) {
			gtdpu_info(chip->dev, DPU_UT_DP, "Trigger the pull-out interrupt and "
					"listen for a real pull-out interrupt.\n");
			queue_delayed_work(gt_dp->irq_event_wq, &gt_dp->irq_event_work, gt_msecs_to_jiffies(1000));
		} else {
			if (dp_get_hpg_status(chip) != 0x1) {
				gt_memset(&gt_dp->native_mode, 0, sizeof(gt_dp->native_mode));
			}

			changed = gt_dp_drm_helper_hpd_irq_event(gt_dp->drm_dev);
			gtdpu_info(chip->dev, DPU_UT_DP,
					"[hpd work]callback drm_helper_hpd_irq_event,"
					"uevent_report:%s\n", changed ? "true" : "false");
#ifdef CONFIG_DRM_GT_AUDIO
			if (!gt_dp->connected)
				gtdpu_connector_clear_eld(&gt_dp->connector);
#endif
		}

		/*If the crtc bound by the current encoder is active after generation the unplug event,
		 * the display link status is read, and of the link loses connection, the link capacity of
		 * the display is compared before and after the unplug, and if the capacity is the same,
		 * the link interaction is carried out to avoid the  upper layer not calling se_crtc resulting
		 * in the inability to display.*/

		if (gt_dp->encoder.crtc &&
			gt_dp->encoder.crtc->state &&
			gt_dp->encoder.crtc->state->active &&
			dp_get_hpg_status(chip)) {
			dp_set_poll_cnt(gt_dp, 3);
			queue_delayed_work(gt_dp->poll_wq, &gt_dp->dp_poll_work, 0);
		}
	}
}

static s32 gt_dp_get_edid_block(void *data, u8 * buf, u32 block, size_t len)
{
	struct dp_chip_t *chip = (struct dp_chip_t*)data;

	if (len > GTDP_EDID_BUF_LEN / 2) {
		return -EINVAL;
	}

	if (block % 2 == 0) {
		gt_memcpy(buf, chip->edid_buf, len);
	} else {
		gt_memcpy(buf, chip->edid_buf + GTDP_EDID_BUF_LEN / 2, len);
	}

	return 0;
}

static int gt_dp_get_modes_strpush(struct drm_connector *connector)
{
	struct dp_device_t *dp_dev = to_dp_device(connector);
	struct dp_chip_t *chip = &dp_dev->chip;
	int ret = 0;

	ret = hal_dp_edid_data(gt_dev_get_parent(chip->dev),
			dp_dev->dp_id, chip->edid_buf);
	if (ret) {
		gtdpu_err(chip->dev, "strpush edid error, ret:%d\n", ret);
		return -EFAULT;
	}

	chip->modes = gtdpu_str_push_edid(chip->edid_buf, connector);

	return ret;
}

static int gt_dp_get_edid_user(struct drm_connector *connector)
{
	struct dp_device_t *dp_dev = to_dp_device(connector);
	struct dp_chip_t *chip = &dp_dev->chip;
	int ret = 0;
	int i = 0;

	ret = hal_dp_edid_data(gt_dev_get_parent(chip->dev),
			dp_dev->dp_id, chip->edid_buf);
	if (ret) {
		gtdpu_err(chip->dev, "get edid from user error, ret:%d\n", ret);
		return -EFAULT;
	}

	for (i = 0; i <= chip->edid_buf[0x7e] && i <= 1; i++) {
		if (!drm_edid_block_valid(chip->edid_buf + i * GTDP_EDID_BUF_LEN / 2, i, false, NULL)) {
			ret = -EFAULT;
			break;
		}
	}

	return ret;
}

int gt_dp_get_edid_monitor(struct drm_connector *connector)
{
	u8 offset = 0;
	struct dp_device_t *dp_dev = to_dp_device(connector);
	int block_valid = 0, retry = 0, ret = 0, i = 0, extension = 0;
	struct i2c_msg msgs[] = {
		{DDC_EDID_ADDR, 0, 1, &offset},
		{DDC_EDID_ADDR, I2C_M_RD, GTDP_EDID_BUF_LEN, dp_dev->chip.edid_buf},
	};

	gt_mutex_lock(dp_dev->aux_mutex);

	for (retry = 0; retry < 4; retry++) {
		gt_memset(dp_dev->chip.edid_buf, 0, sizeof(dp_dev->chip.edid_buf));
		if (i2c_transfer(&dp_dev->aux.ddc, msgs, GT_ARRAY_SIZE(msgs)) == GT_ARRAY_SIZE(msgs)) {
			block_valid = 0;
			extension = (dp_dev->chip.edid_buf[0x7e] >= 1) ? 1 : 0;
			for (i = 0; i <= extension; i++) {
				if (drm_edid_block_valid(dp_dev->chip.edid_buf + i * GTDP_EDID_BUF_LEN / 2, i, false, NULL))
					block_valid++;
				else {
					gtdpu_info(dp_dev->dev, DPU_UT_DP, "read edid failed!"
							"retry:%d block[%d]\n", retry, i);
					if (s_dpu_debug & DPU_UT_DP) {
						gt_print_hex_dump(KERN_NOTICE, " \t", 0, 16, 1,
								dp_dev->chip.edid_buf, GTDP_EDID_BUF_LEN, false);
					}
					ret = -EFAULT;
					break;
				}
			}

			if (block_valid == extension + 1) {
				gtdpu_info(dp_dev->dev, DPU_UT_DP, "read edid success!retry=%d\n", retry);
				ret = 0;
				break;
			}
		}
	}

	gt_mutex_unlock(dp_dev->aux_mutex);

	return ret;
}

static int gt_dp_get_edid(struct drm_connector *connector, int edid_mode)
{
	struct dp_device_t *dp_dev =  to_dp_device(connector);
	struct dp_chip_t *chip = &dp_dev->chip;
	int ret = 0;

	gt_memset(chip->edid_buf, 0, sizeof(chip->edid_buf));

	switch (edid_mode) {
	case EDID_AUTO_READ:
		ret = gt_dp_get_edid_monitor(connector);
	break;
	case EDID_STR_PUSH:
		ret = gt_dp_get_modes_strpush(connector);
	break;
	case EDID_USER_DEFINE:
		ret = gt_dp_get_edid_user(connector);
	break;
	default:
		gtdpu_err(dp_dev->dev, "get_edid failed, invalid edid_mode,"
				"read edid from monitor\n");
		edid_mode = EDID_AUTO_READ;
		ret = -EFAULT;
	}

	gtdpu_info(dp_dev->dev, DPU_UT_DP, "edid_mode:%d ret = %d\n",
			edid_mode, ret);

	if (ret == 0 && edid_mode != EDID_STR_PUSH) {
		dp_set_edid(chip, drm_do_get_edid(connector, gt_dp_get_edid_block, chip));
		chip->max_clk = gtdpu_conn_get_monitor_max_clk((u8 *)chip->edid_buf);
		if (chip->max_clk <= 0) {
			chip->max_clk = 605000;
			gtdpu_info(dp_dev->dev, DPU_UT_DP, "[BAD]get dp max clock err, set it to 605MHz\n");
		}
	}

	if (ret || (edid_mode != EDID_STR_PUSH && dp_get_edid(chip) == NULL)) {
		if (dp_get_edid(chip)) {
			gt_kfree(dp_get_edid(chip));
			dp_set_edid(chip, NULL);
		}
		if (ret != DP_AUX_NATIVE_REPLY_NACK && memchr_inv(chip->edid_buf, 0, GTDP_EDID_BUF_LEN))
			gtdpu_err(dp_dev->dev, "[%s] Edid Invalid, may be poor contact, please re-plug\n", connector->name);
		drm_connector_update_edid_property(connector, NULL);
		dp_set_audio_status(&dp_dev->chip, 0);
		dp_dev->chip.modes = gtdpu_add_modes_without_edid(connector, NULL);
	} else if (edid_mode != EDID_STR_PUSH) {
		drm_connector_update_edid_property(connector, dp_get_edid(&dp_dev->chip));
		dp_set_audio_status(&dp_dev->chip, drm_detect_monitor_audio(dp_get_edid(&dp_dev->chip)));
		dp_dev->chip.modes = drm_add_edid_modes(connector, dp_get_edid(&dp_dev->chip));
	}

	return ret;
}

static bool gt_dp_encoder_mode_fixup(struct drm_encoder *encoder,
		const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct dp_device_t *gt_dp = to_dp_device(encoder);
	const struct drm_display_mode *rmode = NULL;

	if (is_native_mode_valid(&gt_dp->native_mode) && is_virtual_mode(&gt_dp->native_mode, adjusted_mode)) {
		drm_mode_copy(adjusted_mode, &gt_dp->native_mode);
	}

	if (gt_dp->chip.replace_timing) {
		rmode = gtdpu_modes_match_replace_table(adjusted_mode, NULL, gt_dp_is_skip_replace);
		if (!rmode)
			goto out;

		if(!is_special_mode(rmode)) {
			gtdpu_modes_replace_timing(adjusted_mode, rmode);
		}

		gtdpu_info(gt_dp->dev, DPU_UT_HDMI,"%s fixup: "DRM_MODE_FMT "\n",
			gt_dp->name, DRM_MODE_ARG(adjusted_mode));
	}

out:
	return true;
}

static void gt_dp_uevent_report_work(struct work_struct *work)
{
	struct delayed_work *irq_event_work = to_delayed_work(work);
	struct dp_device_t *gt_dp = to_dp_device(irq_event_work);

	if (dp_get_hpg_status(&gt_dp->chip) == 0) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "[uevent_report]callback drm_helper_hpd_irq_event\n");
		gt_dp_drm_helper_hpd_irq_event(gt_dp->drm_dev);
	}
}

static void gt_dp_source_link_init(struct dp_device_t *gt_dp)
{
	if (gt_is_err_or_null(gt_dp->chip.dp_source_link_set))
		return;

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "source caps init\n");

	/* Initialise the number of links on the source side,
	 * the link rate, the average effective symbol transmission unit.
	 */
	gt_dp->chip.dp_source_link_set(&gt_dp->chip, &gt_dp->current_mode);
}

static void gt_dp_poll_execute(struct work_struct *work)
{
	u8 link_cfg[GTDP_LINK_CFG_SIZE];
	u8 link_status[DP_LINK_STATUS_SIZE];
	struct delayed_work *dp_poll_work = to_delayed_work(work);
	struct dp_device_t *gt_dp = to_dp_device(dp_poll_work);

	if (!gt_dp)
		return;

	/* if has panel(backlight), Does not retraining */
	if (gt_dp->panel) {
		return;
	}

	gt_mutex_lock(gt_dp->aux_mutex);

	if (dp_get_poll_cnt(gt_dp) <= 0)
		goto unlock;

	gtdpu_info(gt_dp->dev, DPU_UT_DP, "sink status poll[%d]\n", dp_get_poll_cnt(gt_dp));

	if (gt_dp_need_link_train(gt_dp)) {
		if (gt_dp->chip.dp_source_link_get) {
			gt_dp->chip.dp_source_link_get(&gt_dp->chip, link_cfg);
			gt_dp_caps_prepare(gt_dp);
			gt_dp_rate_adjust(gt_dp, dp_get_poll_cnt(gt_dp) <= 2);
			if (link_cfg[GTDP_LINK_COUNT] != gt_dp->chip.lane_count ||
				link_cfg[GTDP_LINK_RATE]  != gt_dp->chip.lane_rate ||
				dp_get_poll_cnt(gt_dp) <= 2) {
				gt_dp_source_link_init(gt_dp);
			}
		}

		if (gt_dp_rate_valid(gt_dp->chip.lane_rate, gt_dp->chip.lane_count,
			gt_dp->current_mode.clock, gt_dp->chip.display_info.bpc)) {
			if (!gt_dp_clock_recover(gt_dp))
				goto fail_handle;
			if (!gt_dp_channel_balance(gt_dp))
				goto fail_handle;
			gt_dp_link_start(gt_dp);
			if (gt_dp_need_link_train(gt_dp) && (dp_get_poll_cnt(gt_dp) == 2) &&
				memchr_inv(gt_dp->chip.edid_buf, 0, GTDP_EDID_BUF_LEN)) {
				gtdpu_err(gt_dp->chip.dev, "Link invalid!\n");
			}
			gt_dp_update_audio(gt_dp);
		}
	}

fail_handle:
	dp_set_poll_cnt(gt_dp, dp_get_poll_cnt(gt_dp) - 1);
	if (dp_get_poll_cnt(gt_dp) >= 1) {
		queue_delayed_work(gt_dp->poll_wq, &gt_dp->dp_poll_work, gt_msecs_to_jiffies(DP_POLL_PERIOD_MS));
	} else if (gt_dp_monitor_en(dp_link_monitor) && dp_get_edid(&gt_dp->chip) &&
		(sink_lane_status_get(gt_dp, link_status) != DP_AUX_NATIVE_REPLY_NACK)) {
		dp_set_poll_cnt(gt_dp, 1);
		queue_delayed_work(gt_dp->poll_wq, &gt_dp->dp_poll_work, gt_msecs_to_jiffies(DP_POLL_PERIOD_MS * 200));
	}

unlock:
	gt_mutex_unlock(gt_dp->aux_mutex);
}

static void gt_dp_encoder_mode_set(struct drm_encoder *encoder,
		struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	int dpu_id = 0;
	struct dp_device_t *gt_dp = to_dp_device(encoder);

	if (!gt_dp || !adjusted_mode)
		return;

	dpu_id = gtdpu_get_dpuid_bycrtc(encoder->crtc);

	drm_mode_copy(&gt_dp->current_mode, adjusted_mode);

	gt_dp_caps_prepare(gt_dp);

	gt_dp_video_timing_prepare(gt_dp);

	if (gt_dp->chip.encoder_modeset) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP,"%s modeset: "DRM_MODE_FMT "\n",
			gt_dp->name, DRM_MODE_ARG(adjusted_mode));
		gt_dp->chip.encoder_modeset(&gt_dp->chip, dpu_id, false, adjusted_mode);
	}
}

#ifdef CONFIG_DRM_GT_AUDIO
static void gt_dp_audio_enable(struct audio_conn *ac)
{
	struct dp_device_t *gt_dp = NULL;

	if (!ac || !ac->priv) {
		return;
	}

	gt_dp = (struct dp_device_t *)ac->priv;

	if (gt_dp->chip.dp_audio_enable &&
		dp_get_audio_status(&gt_dp->chip)) {
		gt_dp->audio_enable = true;
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "audio enable\n");
		gt_dp->chip.dp_audio_enable(&gt_dp->chip, ac->data.sample_bits);
	}
}

static void gt_dp_audio_disable(struct audio_conn *ac)
{
	struct dp_device_t *gt_dp = NULL;

	if (!ac || !ac->priv) {
		return;
	}

	gt_dp = (struct dp_device_t *)ac->priv;

	if (!gt_dp->audio_enable)
		return;

	if (gt_dp->chip.dp_audio_disable) {
		gt_dp->audio_enable = false;
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "audio disable\n");
		gt_dp->chip.dp_audio_disable(&gt_dp->chip);
	}
}

#else
static inline void gt_dp_audio_enable(struct audio_conn *ac){}
static inline void gt_dp_audio_disable(struct audio_conn *ac){}
#endif

static void gt_dp_encoder_mode_disable(struct drm_encoder *encoder)
{
	struct dp_device_t *gt_dp = to_dp_device(encoder);

	if (gt_is_err_or_null(gt_dp))
		return;

	gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"gt %s encoder mode disable\n",gt_dp->name);

	gt_mutex_lock(gt_dp->aux_mutex);

	gt_dp_sink_power_ctrl(gt_dp, false);

	/* Backlight off */
	gt_panel_disable(gt_dp->panel);

	/* Source
	 * 1. Main Tx disabled
	 * */
	if (gt_dp->chip.encoder_disable)
		gt_dp->chip.encoder_disable(&gt_dp->chip);

	/*notification of audio devices */
	gt_dp_update_audio(gt_dp);

	dp_set_poll_cnt(gt_dp, 0);
	gt_mutex_unlock(gt_dp->aux_mutex);

	cancel_delayed_work_sync(&gt_dp->dp_poll_work);

	/* Switch off the power */
	gt_panel_unprepare(gt_dp->panel);
}

static void gt_dp_encoder_mode_enable(struct drm_encoder *encoder)
{
	struct dp_device_t *gt_dp = to_dp_device(encoder);

	if (gt_is_err_or_null(gt_dp))
		return;

	gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"gt %s encoder mode enable\n", gt_dp->name);

	gt_mutex_lock(gt_dp->aux_mutex);
	gt_dp_sink_power_ctrl(gt_dp, true);
	gt_dp_clock_recover(gt_dp);
	gt_dp_channel_balance(gt_dp);

	/*Enable video stream-dp_device*/
	if (gt_dp->chip.encoder_enable)
		gt_dp->chip.encoder_enable(&gt_dp->chip);
	if (gt_dp->is_R1) {
		gt_panel_enable(gt_dp->panel);
	}
	gt_dp_link_start(gt_dp);

	if (gt_dp->audio_enable)
		gt_dp_audio_enable(gt_dp->ac);

	gt_dp_update_audio(gt_dp);

	gt_mutex_unlock(gt_dp->aux_mutex);

	gt_panel_enable(gt_dp->panel);

	dp_set_poll_cnt(gt_dp, 5);
	queue_delayed_work(gt_dp->poll_wq, &gt_dp->dp_poll_work,
			gt_msecs_to_jiffies(DP_POLL_PERIOD_MS));
}

static void gt_dp_encoder_destroy(struct drm_encoder *encoder)
{
	if (gt_is_err_or_null(encoder))
		return;

	drm_encoder_cleanup(encoder);
}

static void gt_dp_connector_destroy(struct drm_connector *connector)
{
	if (gt_is_err_or_null(connector))
		return;

	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static void gt_dp_cleanup_connector_encoder(
			struct dp_device_t *gt_dp)
{
	/*
	 * After the dp reference count is zeroed, drm_connector_free will be called
	 * encoder autorelease function without reference counting
	 */
	gt_dp_connector_destroy(&gt_dp->connector);
	gt_dp_encoder_destroy(&gt_dp->encoder);
}

static enum drm_mode_status gt_dp_connector_helper_mode_valid(struct drm_connector *connector,
					   struct drm_display_mode *mode)
{
	gt_drm_mode_status status = MODE_OK;
	struct dp_device_t *gt_dp = to_dp_device(connector);

	if (gt_is_err_or_null(gt_dp))
		return MODE_OK;

	if (!gt_dp_rate_valid(gt_dp->chip.max_source_rate, 4, mode->clock, gt_dp_bpc[1])) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP,
				"[BAD]clock required is too high,""[%dx%d clock:%d]kHZ\n",
				mode->hdisplay, mode->vdisplay, mode->clock);
		return MODE_CLOCK_HIGH;
	}

	if (gt_dp->chip.connector_mode_valid)
		status = gt_dp->chip.connector_mode_valid(&gt_dp->chip, mode);

	if (status != MODE_OK) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP,
			"%s connector mode invalid: %dx%d@%d valid:%d",
			gt_dp->name, mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), status);
	}

	return status;
}

static bool gt_dp_native_mode_filter(struct drm_connector *connector,
	struct drm_display_mode *mode, bool priv_filter)
{
	struct dp_device_t *gt_dp = to_dp_device(connector);
#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	struct drm_encoder *encoder = &gt_dp->encoder;
#endif
	int ret = 0;

	// TBD: current not support 4k scaler
	if (mode && priv_filter && ((mode->hdisplay > GTDPU_COMBINE_WIDTH) ||
		(mode->vdisplay > GTDPU_COMBINE_HEIGHT)) )
		return true;

	if (mode && gt_dp->chip.max_clk > 0 && \
		mode->clock > gt_dp->chip.max_clk)
		return true;

	if (connector->helper_private && connector->helper_private->mode_valid) {
		ret = connector->helper_private->mode_valid(&gt_dp->connector, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}

#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	if (encoder->helper_private && encoder->helper_private->mode_valid) {
		ret = encoder->helper_private->mode_valid(&gt_dp->encoder, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}
#endif

	return false;
}

static void gt_dp_common_mode_add(struct drm_connector * connector)
{
	struct drm_display_mode *native_mode = NULL;
	struct dp_device_t *gt_dp = to_dp_device(connector);

	if (gt_is_err_or_null(gt_dp))
		return;

	// get native mode
	drm_mode_sort(&connector->probed_modes);
	native_mode = gtdpu_get_native_mode(connector, gt_dp_native_mode_filter);

	// add common mode
	if (native_mode) {
		gtdpu_info(gt_dp->dev, DPU_UT_DP, "%s native mode: "DRM_MODE_FMT ", status:%d\n",
			gt_dp->name, DRM_MODE_ARG(native_mode), native_mode->status);
		gt_memcpy(&gt_dp->native_mode, native_mode, sizeof(gt_dp->native_mode));
		drm_mode_set_crtcinfo(&gt_dp->native_mode, CRTC_INTERLACE_HALVE_V);
		gt_dp->native_mode.status = MODE_OK;
		gt_dp->chip.modes += gtdpu_connector_add_common_modes(connector, &gt_dp->native_mode, false);
	}
}

static int gt_dp_connector_helper_get_modes(struct drm_connector *connector)
{
	struct dp_device_t *gt_dp = to_dp_device(connector);

	if (gt_is_err_or_null(gt_dp))
		return 0;

	gt_memset(&gt_dp->native_mode, 0, sizeof(gt_dp->native_mode));
	if (dp_get_edid(&gt_dp->chip)) {
		drm_connector_update_edid_property(connector, dp_get_edid(&gt_dp->chip));
		dp_set_audio_status(&gt_dp->chip, drm_detect_monitor_audio(dp_get_edid(&gt_dp->chip)));
		gt_dp->chip.modes = drm_add_edid_modes(connector, dp_get_edid(&gt_dp->chip));
	}

	if (!dp_get_edid(&gt_dp->chip) || (gt_dp->chip.hal_edid_mode == EDID_STR_PUSH))
		gt_dp_get_edid(connector, gt_dp->chip.hal_edid_mode);

	if (gt_dp->chip.hal_zoom_enable && !dp_get_edid(&gt_dp->chip))
		gt_dp_get_edid(connector, EDID_AUTO_READ);

#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
	drm_edid_to_eld(connector, dp_get_edid(&gt_dp->chip));
#endif

	gt_dp->chip.modes -= gtdpu_modes_drop_repeat(connector);

	// fixup 1366x768 and more modes for /dev/fb0
	gtdpu_modes_fixup_preferred_nonaligned_modes(connector);

	if (s_dpu_debug & DPU_UT_DP) {
		gt_print_hex_dump(KERN_NOTICE, " \t", 0, 16, 1,
			gt_dp->chip.edid_buf, GTDP_EDID_BUF_LEN, false);
	}

	gt_dp_common_mode_add(connector);

	return gt_dp->chip.modes;
}

static ssize_t
gt_dp_aux_transfer(struct drm_dp_aux *aux, struct drm_dp_aux_msg *msg)
{
	int ret = 0, j = 0;
	struct aux_cfg *aux_vcp = NULL;
	struct dp_device_t *dp = to_dp_device(aux);

	if (gt_is_err_or_null(dp))
		return -EFAULT;

	aux_vcp = gt_kzalloc_kernel(sizeof(struct aux_cfg));
	if (gt_is_err_or_null(aux_vcp))
		return -ENOMEM;

	switch (msg->request & ~DP_AUX_I2C_MOT) {
	case DP_AUX_NATIVE_WRITE:
	case DP_AUX_I2C_WRITE:
	case DP_AUX_I2C_WRITE_STATUS_UPDATE:
	{
		aux_vcp->aux_cmd  = msg->request;
		aux_vcp->dpcd_addr = msg->address;
		aux_vcp->length = msg->size ? (msg->size - 1) : 0x10;
		aux_vcp->read  = 0;
		if (msg->size) {
			for (j = 0; j < msg->size; j++)
				aux_vcp->wr_buff[j / 4] |= (((u8 *)msg->buffer)[j] << ((j % 4) * 8));
		}

		ret |= gt_dp_aux_rw(dp, aux_vcp);
	}
	break;
	case DP_AUX_NATIVE_READ:
	case DP_AUX_I2C_READ:
	{
		aux_vcp->aux_cmd = msg->request;
		aux_vcp->dpcd_addr = msg->address;
		aux_vcp->length = msg->size ? (msg->size - 1) : 0x10;
		aux_vcp->read = 1;
		ret |= gt_dp_aux_rw(dp, aux_vcp);
		for (j = 0; j < msg->size; j++)
			((u8 *)msg->buffer)[j] = (aux_vcp->rd_buff[j / 4] >> (8 *(j % 4))) & 0xff;
	}
	break;
	default:
		ret = -EINVAL;
		break;
	}

	msg->reply = ret;

	if (msg->reply & DP_AUX_NATIVE_REPLY_DEFER ||
		msg->reply & DP_AUX_I2C_REPLY_DEFER) {
		dp_info(dp->dev, "[Note]Aux Defer:%d\n", msg->reply);
	}

	gt_kfree(aux_vcp);

	return msg->size;
}

static const struct drm_connector_funcs s_gtdp_connector_funcs = {
	.dpms       = drm_helper_connector_dpms,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = gt_dp_connector_destroy,
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
	.late_register = gt_dp_late_register,
	.early_unregister = gt_dp_early_unregister,
	.detect = gt_dp_detect,
};

static struct drm_connector_helper_funcs s_gtdp_connector_helper_funcs = {
	.get_modes = gt_dp_connector_helper_get_modes,
	.mode_valid = gt_dp_connector_helper_mode_valid,
#if ((DRM_VERSION >= KERNEL_VERSION(4, 12, 0)))
	.detect_ctx = gt_dp_detect_ctx,
#endif
};

static const struct drm_encoder_helper_funcs s_gtdp_encoder_helper_funcs = {
	.mode_fixup = gt_dp_encoder_mode_fixup,
	.mode_set = gt_dp_encoder_mode_set,
	.disable = gt_dp_encoder_mode_disable,
	.enable = gt_dp_encoder_mode_enable,
};

static const struct drm_encoder_funcs s_gtdp_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static int gt_dp_connector_create(struct drm_device *dev, struct drm_connector *connector)
{
	int connector_type = 0;
	struct dp_device_t * dp_dev = to_dp_device(connector);

	if (NULL == dev || NULL == connector) {
		DRM_ERROR("Invalid argument: the pointer of dev is %p, connector is %p!\n", dev, connector);
		return -EFAULT;
	}

	connector->dpms = DRM_MODE_DPMS_OFF;
	connector->polled = DRM_CONNECTOR_POLL_HPD;

	if (dp_dev->chip.output_mode) {
		connector_type = dp_dev->chip.output_mode->mode_connector_item;
	} else {
		connector_type = DRM_MODE_CONNECTOR_DisplayPort;
	}

	drm_connector_init(dev, connector, &s_gtdp_connector_funcs, connector_type);
	drm_connector_helper_add(connector, &s_gtdp_connector_helper_funcs);

	return 0;
}

static int gt_dp_encoder_create(struct drm_device *drm_dev,
		struct drm_encoder *encoder, unsigned int possible_crtc)
{
	int retcode = 0;

	retcode = drm_encoder_init(drm_dev, encoder, &s_gtdp_encoder_funcs, DRM_MODE_ENCODER_TMDS, NULL);
	if (retcode) {
		DRM_ERROR("Failed to initialise DP encoder\n");
		return retcode;
	}

	drm_encoder_helper_add(encoder, &s_gtdp_encoder_helper_funcs);
	encoder->possible_crtcs = possible_crtc;

	return retcode;
}

static int gtdp_encoder_connector_attach(struct drm_device *drm_dev, struct dp_device_t *gt_dp)
{
	int ret = 0;

	if (NULL == drm_dev || NULL == gt_dp) {
		DRM_ERROR("Invalid argument: drm_dev or dp is NULL!\n");
		return -1;
	}

	ret = gt_dp_encoder_create(drm_dev, &gt_dp->encoder, gt_dp->chip.possible_crtc);
	if (ret) {
		gtdpu_err(gt_dp->dev, " DP gtdp_encoder_create failed!\n");
		goto err_config_cleanup;
	}

	ret = gt_dp_connector_create(drm_dev, &gt_dp->connector);
	if (ret) {
		gtdpu_err(gt_dp->dev, "DP gtdp_connector_create failed!\n");
		goto err_config_cleanup;
	}

	ret = drm_connector_attach_encoder(&gt_dp->connector, &gt_dp->encoder);

	gtdpu_info(drm_dev->dev, DPU_UT_DP, "dp attach [ENCODER:%d] to [CONNECTOR:%d] (ret = %d)\n",
				gt_dp->encoder.base.id, gt_dp->connector.base.id, ret);
	goto attach_exit;

err_config_cleanup:
	drm_mode_config_cleanup(drm_dev);
attach_exit:
	return ret;
}

static int gtdpu_dp_zoomenable_id(int dp_id)
{
	int zoom_id = -1;

	switch (dp_id) {
		case 0:
			zoom_id = DP0_ZOOM_ENABLE;
			break;
		case 1:
			zoom_id = DP1_ZOOM_ENABLE;
			break;
		default:
			break;
	}
	return zoom_id;
}

static int gt_dp_chip_init(struct dp_device_t *gt_dp)
{
	chip_type_e plat;
	int retcode = 0;
	int zoom_id = 0;

	gt_dp->chip.lane_rate = 0x14;
	gt_dp->chip.lane_count = 0x04;
	gt_dp->chip.phy_lanes = 0x2;
	gt_dp->chip.phy_rate = 0x1;
	gt_dp->chip.enhance_mode = 0x01;
	gt_dp->chip.lane_swing[0] = 0x2;
	gt_dp->chip.lane_swing[1] = 0x2;
	gt_dp->chip.lane_swing[2] = 0x2;
	gt_dp->chip.lane_swing[3] = 0x2;
	gt_dp->chip.lane_emphasis[0] = 0x03 << DP_TRAIN_PRE_EMPHASIS_SHIFT;
	gt_dp->chip.lane_emphasis[1] = 0x03 << DP_TRAIN_PRE_EMPHASIS_SHIFT;
	gt_dp->chip.lane_emphasis[2] = 0x03 << DP_TRAIN_PRE_EMPHASIS_SHIFT;
	gt_dp->chip.lane_emphasis[3] = 0x03 << DP_TRAIN_PRE_EMPHASIS_SHIFT;
	gt_dp->chip.dp_blight_value = 90;

	gt_dp->chip.hal_edid_mode = hal_dp_edid_mode(gt_dp->parent, gt_dp->dp_id);
	if (gt_dp->chip.hal_edid_mode < EDID_AUTO_READ) {
		dp_info(gt_dp->dev, "hal_dp-%d edid_mode parse error%d\n",
				gt_dp->dp_id, gt_dp->chip.hal_edid_mode);
		gt_dp->chip.hal_edid_mode = EDID_AUTO_READ;
	}

	zoom_id = gtdpu_dp_zoomenable_id(gt_dp->dp_id);
	if ((zoom_id >= 0) && (hal_zoom_is_enable(gt_dp->parent, zoom_id) == 1)) {
		gt_dp->chip.hal_zoom_enable = true;
		gt_dp->zoom_id = zoom_id;
	} else {
		gt_dp->chip.hal_zoom_enable = false;
		gt_dp->zoom_id = 0;
	}

	plat = hal_get_chiptype(gt_dp->parent);
	switch(plat) {
	case CHIP_GT6912_SOC:
		gt_dp->chip.hal_zoom_enable = false;
		retcode = gt6912_soc_dp_chip_init(&gt_dp->chip,
				gt_dp->dev, gt_dp->dp_id);
		break;
	case CHIP_GT6910_SOC:
		retcode = gt6910_soc_dp_chip_init(&gt_dp->chip,
				gt_dp->dev, gt_dp->dp_id);
		break;
	case CHIP_GT6913_SOC:
		gt_dp->chip.hal_zoom_enable = false;
		retcode = gt6913_soc_dp_chip_init(&gt_dp->chip,
				gt_dp->dev, gt_dp->dp_id);
		break;
	case CHIP_GT6908_SOC:
		retcode = gt6908_soc_dp_chip_init(&gt_dp->chip,
				gt_dp->dev, gt_dp->dp_id);
		break;
	default:
		gtdpu_err(gt_dp->dev, "%s does not currently support %d platform.\n",
			gt_dp->name, plat);
		retcode = -EINVAL;
		break;
	}

	gt_dp->chip.output_mode = gtdpu_get_connector_output_mode(gt_dp->dev, REG_M_DP);

	if (gt_dp->chip.output_mode) {
		if (gt_dp->chip.output_mode->mode_connector_item == DRM_MODE_CONNECTOR_VGA) {
			gt_dp->chip.max_width = 1920;
			gt_dp->chip.max_height = 1080;
		}
		dp_info(gt_dp->dev, "mode_connector:%d convert:%s",
				gt_dp->chip.output_mode->mode_connector_item, gt_dp->chip.output_mode->convert_name);
	}

	dp_info(gt_dp->dev, "hal_dp-%d zoom:%s\n",
			gt_dp->dp_id, gt_dp->chip.hal_zoom_enable ? "enable" : "disable");
	dp_info(gt_dp->dev, "hal_dp-%d edid_mode %d\n",
			gt_dp->dp_id, gt_dp->chip.hal_edid_mode);

	return retcode;
}

static void gt_dp_chip_fini(struct dp_device_t *gt_dp)
{
	chip_type_e plat;

	plat = hal_get_chiptype(gt_dp->parent);
	switch(plat) {
	case CHIP_GT6912_SOC:
		gt6912_soc_dp_chip_fini(&gt_dp->chip);
		break;
	case CHIP_GT6910_SOC:
		gt6910_soc_dp_chip_fini(&gt_dp->chip);
		break;
	case CHIP_GT6913_SOC:
		gt6913_soc_dp_chip_fini(&gt_dp->chip);
		break;
	case CHIP_GT6908_SOC:
		gt6908_soc_dp_chip_fini(&gt_dp->chip);
		break;
	default:
		gtdpu_err(gt_dp->dev, "%s does not currently support %d platform.\n",
			gt_dp->name, plat);
		break;
	}
}

static void gt_dp_audio_register(struct dp_device_t *gt_dp)
{
#ifdef CONFIG_DRM_GT_AUDIO
	struct audio_conn *pac = NULL;

	if (hal_getflag_dp2vga(gt_dp->parent, gt_dp->dp_id) <= 0) {
		pac = kzalloc(sizeof(struct audio_conn), gt_gfp_kernel);
		if (!pac) {
			gtdpu_err(gt_dp->dev, "Alloc audio_conn failed.\n");
			return;
		}

		pac->conn_st = 0;
		pac->dev = (void *)gt_dp->dev;
		pac->id = gt_dp->dp_id;
		pac->type = GTAUDIO_CONNECTOR_TYPE_DP;
		pac->has_audio = 1; //todo
		pac->priv = (void *)gt_dp;
		pac->enable  = gt_dp_audio_enable;
		pac->disable = gt_dp_audio_disable;
		gt_dp->ac = pac;
		if (gtaudio_register_connector(pac)) {
			dp_info(gt_dp->dev, "register audio_conn failed.\n");
			gt_dp->ac = NULL;
			gt_kfree(pac);
		} else {
			dp_info(gt_dp->dev, "register audio_conn ok.\n");
			gt_dp->ac = pac;
		}
	}
#endif
}

static int gtdpu_dp_bind(struct device *dev, struct device *master, void *data)
{
	int retcode = 0;
	struct drm_device * drm_dev = data;
	struct dp_device_t *gt_dp = NULL;
	plat_data_t *pdata = gt_dev_get_platdata(dev);

	BUG_ON(!dev);
	BUG_ON(!data);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_DP0 + pdata->dev_idx)) {
		dp_info(dev, "possible_crtc = 0, do not bind dp-%d\n", pdata->dev_idx);
		return retcode;
	}

	gt_dp = gt_devm_kmalloc_kernel(dev, sizeof(*gt_dp));
	if (!gt_dp) {
		gtdpu_err(dev, "Alloc dp handle failed. Short of memory.\n");
		return -ENOMEM;
	}
	gt_memset(gt_dp, 0, sizeof(*gt_dp));
	gt_dp->dp_id = pdata->dev_idx;
	gt_dp->dev = gt_get_device(dev);
	gt_dp->parent = gt_dev_get_parent(dev);
	gt_dp->drm_dev = drm_dev;
	gt_dp->name = gt_kasprintf(gt_gfp_kernel, "gt-dp-%d",
		gt_dp->dp_id);
	if (!gt_dp->name) {
		gtdpu_err(dev, "Alloc dp-%d name failed. Short of memory.\n",
			pdata->dev_idx);
		retcode = -ENOMEM;
		goto err_out_name;
	}
	gt_dp->aux_msg.buffer = gt_kzalloc_kernel(GTDP_EDID_BUF_LEN);
	if (!gt_dp->aux_msg.buffer) {
		gtdpu_err(dev, "Alloc aux_msg failed. Short of memory.\n");
		retcode = -ENOMEM;
		goto err_out_aux;
	}

	dev_set_drvdata(dev, gt_dp);

	retcode = gt_dp_chip_init(gt_dp);
	if (retcode) {
		gtdpu_err(dev, "%s chip Init failed-%d.\n", gt_dp->name, retcode);
		goto err_chip_init;
	}

	retcode = gtdp_encoder_connector_attach(drm_dev, gt_dp);
	if (retcode) {
		gtdpu_err(gt_dp->dev, "gtdp_encoder_connector_attach failed, ret-%d!\n", retcode);
		goto err_attach_encoder;
	}

	gt_dp->aux_mutex = gt_mutex_alloc();
	gt_dp->hpd_wq = gt_create_singlethread_wq("dp hpd_wq");
	gt_dp->hpd_work = gt_work_alloc(gt_dp_hpd_work, (void *)gt_dp);
	gt_dp->poll_wq = gt_create_singlethread_wq("dp poll_wq");
	gt_dp->irq_event_wq = gt_create_singlethread_wq("dp irq_event_wq");
	if (gt_is_err_or_null(gt_dp->hpd_wq)   ||
		gt_is_err_or_null(gt_dp->hpd_work) ||
		gt_is_err_or_null(gt_dp->poll_wq)  ||
		gt_is_err_or_null(gt_dp->irq_event_wq)) {
		gtdpu_err(gt_dp->dev, "gtdp hpd_work create failed!\n");
		goto err_hpd_work;
	}
	INIT_DELAYED_WORK(&gt_dp->dp_poll_work, gt_dp_poll_execute);
	INIT_DELAYED_WORK(&gt_dp->irq_event_work, gt_dp_uevent_report_work);

	retcode = hal_set_irq_handler(dev->parent, gt_dp->chip.hal_module, gt_dp_irq_handle, &gt_dp->chip);
	if (retcode) {
		gtdpu_err(dev, "failed to set interrupt handler (err = %d)\n", retcode);
		goto err_enable_irq;
	}

	retcode = gt_dp_hw_init(gt_dp);
	if (retcode) {
		gtdpu_err(dev, "%s hw Init failed-%d.\n", gt_dp->name, retcode);
		goto err_hw_init;
	}

	hal_dev_enable_irq(dev->parent, gt_dp->chip.hal_module);
	gt_dp_hw_irq_enable(gt_dp);

#if (DRM_VERSION >= KERNEL_VERSION(5, 14, 0))
	gt_dp->aux.drm_dev = drm_dev;
	gt_dp->aux.name = kasprintf(gt_gfp_kernel, "gt-dp[%d] aux", pdata->dev_idx);
#endif
	drm_dp_aux_init(&gt_dp->aux);
	gt_dp->aux.dev = gt_dp->dev;
	gt_dp->aux.transfer = gt_dp_aux_transfer;

	gt_dp_audio_register(gt_dp);

	return retcode;

err_hpd_work:
err_enable_irq:
	gt_dp_cleanup_connector_encoder(gt_dp);
err_attach_encoder:
	gt_dp_hw_fini(gt_dp);
err_hw_init:
	gt_dp_chip_fini(gt_dp);
err_chip_init:
	gt_kfree(gt_dp->aux_msg.buffer);
err_out_aux:
	gt_kfree(gt_dp->name);
err_out_name:
	put_device(dev);
	gt_devm_kfree(dev, gt_dp);

	return retcode;
}

static void gtdpu_dp_unbind(struct device *dev, struct device *master, void *data)
{
	struct dp_device_t *gt_dp = NULL;
	plat_data_t *pdata = gt_dev_get_platdata(dev);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_DP0 + pdata->dev_idx)) {
		dp_info(dev, "possible_crtc = 0, do not unbind dp-%d\n", pdata->dev_idx);
		return;
	}

	gt_dp = dev_get_drvdata(dev);
	if (!gt_dp) {
		gtdpu_err(dev, "dp handle is NULL\n");
		return;
	}

	gt_dp_hw_irq_disable(gt_dp);
	hal_dev_disable_irq(dev->parent, gt_dp->chip.hal_module);
	hal_set_irq_handler(dev->parent,
		gt_dp->chip.hal_module, NULL, NULL);

	gt_flush_work(gt_dp->hpd_work);
	gt_work_destroy(gt_dp->hpd_work);
	gt_destroy_workqueue(gt_dp->hpd_wq);
	cancel_delayed_work_sync(&gt_dp->dp_poll_work);
	gt_destroy_workqueue(gt_dp->poll_wq);
	cancel_delayed_work_sync(&gt_dp->irq_event_work);
	gt_destroy_workqueue(gt_dp->irq_event_wq);
	gt_mutex_free(gt_dp->aux_mutex);

	gt_dp_hw_fini(gt_dp);
	gt_dp_chip_fini(gt_dp);
	dev_set_drvdata(dev, NULL);

	if (gt_dp->aux_msg.buffer)
		gt_kfree(gt_dp->aux_msg.buffer);
	if (gt_dp->name)
		gt_kfree(gt_dp->name);
	if (gt_dp->aux.name)
		gt_kfree(gt_dp->aux.name);

#ifdef CONFIG_DRM_GT_AUDIO
	if(gt_dp->ac){
		gtaudio_unregister_connector(gt_dp->ac);
		gt_kfree(gt_dp->ac);
	}
#endif
	put_device(dev);
	gt_devm_kfree(dev, gt_dp);

	return;
}

static const struct component_ops s_gtdpu_dp_ops = {
	.bind = gtdpu_dp_bind,
	.unbind = gtdpu_dp_unbind,
};

static int gtdpu_dp_probe(struct platform_device *pdev)
{
	if (NULL == pdev) {
		DRM_ERROR("Invalid argument: the value of pdev is NULL!\r\n");
		return -1;
	}

	return component_add(&pdev->dev, &s_gtdpu_dp_ops);
}

static int gtdpu_dp_remove(struct platform_device *pdev)
{
	if (NULL == pdev) {
		DRM_ERROR("Invalid argument: the value of pdev is NULL!\r\n");
		return -EFAULT;
	}
	component_del(&pdev->dev, &s_gtdpu_dp_ops);
	return 0;
}

static struct platform_device_id s_gtdpu_dp_platform_device_id_table[] = {
	{.name = GT_DP_DEVICE_NAME,.driver_data = 0},
	{},
};
MODULE_DEVICE_TABLE(platform, s_gtdpu_dp_platform_device_id_table);

static int gtdpu_dp_suspend(struct device *dev)
{
	struct dp_device_t *dp = dev_get_drvdata(dev);
	plat_data_t *pdata = gt_dev_get_platdata(dev);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_DP0 + pdata->dev_idx)) {
		dp_info(dev, "possible_crtc = 0, do not suspend dp-%d\n", pdata->dev_idx);
		return 0;
	}

	if (dp) {
		gtdpu_info(dp->dev, DPU_UT_DP, "suspend\n");

		gt_flush_work(dp->hpd_work);
		cancel_delayed_work_sync(&dp->dp_poll_work);
		cancel_delayed_work_sync(&dp->irq_event_work);

		gt_dp_hw_irq_disable(dp);
		hal_dev_disable_irq(dev->parent, dp->chip.hal_module);

		gt_dp_hw_fini(dp);
	}

	return 0;
}

static int gtdpu_dp_resume(struct device *dev)
{
	struct dp_device_t *dp = dev_get_drvdata(dev);
	plat_data_t *pdata = gt_dev_get_platdata(dev);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_DP0 + pdata->dev_idx)) {
		dp_info(dev, "possible_crtc = 0, do not resume dp-%d\n", pdata->dev_idx);
		return 0;
	}

	if (dp) {
		gtdpu_info(dp->dev, DPU_UT_DP, "resume\n");

		gt_dp_hw_init(dp);
		hal_dev_enable_irq(dev->parent, dp->chip.hal_module);
		gt_dp_hw_irq_enable(dp);
		gt_dp_check_sink_connection(dp);
	}

	return 0;
}

static void gtdpu_dp_shutdown(struct platform_device *pdev)
{
	struct dp_device_t *dp = NULL;
	plat_data_t *pdata =  dev_get_platdata(&pdev->dev);

	if (!gtdpu_detect_is_valid_output(&pdev->dev, CONNECTOR_M_DP0 + pdata->dev_idx)) {
		dp_info(&pdev->dev, "possible_crtc = 0, do not shutdown dp-%d\n", pdata->dev_idx);
		return;
	}

	dp = gt_platform_get_drvdata(pdev);
	if (!dp)
		return;

	gt_dp_encoder_mode_disable(&dp->encoder);
	gtdpu_dp_suspend(&pdev->dev);
}

static const struct dev_pm_ops gtdpu_dp_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(gtdpu_dp_suspend, gtdpu_dp_resume)
};

struct platform_driver g_gtgpu_dp_driver = {
	.probe = gtdpu_dp_probe,
	.remove = gtdpu_dp_remove,
	.shutdown = gtdpu_dp_shutdown,
	.driver = {
		.name = GT_DP_DEVICE_NAME,
		.pm = &gtdpu_dp_pm_ops,
	},
	.id_table = s_gtdpu_dp_platform_device_id_table,
};
