/*************************************************************************/ /*!
@File			gtdpu_hdmi.c
@Title
@Copyright		Copyright (c) Gengtu Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include <linux/platform_device.h>
#include <drm/drm_modes.h>
#include <drm/drm_edid.h>

#include "gtdpu_hdmi.h"
#include "gtdpu_hdmi_debugfs.h"
#include "gtdpu_connector.h"

#ifdef CONFIG_DRM_GT_AUDIO
#include "gtaudio_drv.h"
#endif

static int s_hdmi_hwi2c = true;
module_param(s_hdmi_hwi2c, int, 0600);
MODULE_PARM_DESC(s_hdmi_hwi2c, "hdmi hwi2c mode.Only gt6908 support(default: true)");

static bool s_hdmi_support_75hz = true;
module_param(s_hdmi_support_75hz, bool, 0600);
MODULE_PARM_DESC(s_hdmi_support_75hz, "hdmi support 75hz (default: true)");

static bool gt_hdmi_encoder_mode_fixup(struct drm_encoder *encoder,
		const struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);
	const struct drm_display_mode *rmode = NULL;

	if (is_native_mode_valid(&gt_hdmi->native_mode) && \
			is_virtual_mode(&gt_hdmi->native_mode, adjusted_mode)) {
		drm_mode_copy(adjusted_mode, &gt_hdmi->native_mode);
	}

	if (gt_hdmi->chip.replace_timing) {
		rmode = gtdpu_modes_match_replace_table(adjusted_mode, NULL, NULL);
		if (!rmode)
			goto out;

		if (!is_special_mode(rmode)) {
			gtdpu_modes_replace_timing(adjusted_mode, rmode);
		}

		gtdpu_info(gt_hdmi->dev, DPU_UT_HDMI,"%s fixup: "DRM_MODE_FMT "\n",
			gt_hdmi->name, DRM_MODE_ARG(adjusted_mode));
	}

out:
	return true;
}

static void gt_hdmi_encoder_mode_set(struct drm_encoder *encoder,
		struct drm_display_mode *mode, struct drm_display_mode *adjusted_mode)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);
	int dpu_id = 0;

	if (!encoder || !adjusted_mode) {
		return;
	}

	atomic64_set(&gt_hdmi->modesetting, 1);

	dpu_id = gtdpu_get_dpuid_bycrtc(encoder->crtc);

	if (gt_hdmi->chip.encoder_modeset) {
		hdmi_info(gt_hdmi->dev, "%s modeset: "DRM_MODE_FMT "\n",
			gt_hdmi->name, DRM_MODE_ARG(adjusted_mode));
		drm_mode_copy(gt_hdmi->chip.adjusted_mode, adjusted_mode);
		gt_hdmi->chip.encoder_modeset(&gt_hdmi->chip,dpu_id, gt_hdmi->chip.test_mode, adjusted_mode);
	}

	return;
}

static void gt_hdmi_encoder_mode_disable(struct drm_encoder *encoder)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);

	if (NULL == encoder) {
		return;
	}

	gt_hdmi->scramb_active = 0;

#ifdef CONFIG_DRM_GT_AUDIO
	if(gt_hdmi->ac && gt_hdmi->ac->report_jack){
		int dpms, conn_st;
		hdmi_info(gt_hdmi->dev,
			"%s try report jack, conn_st:%d, has_audio:%d",
			gt_hdmi->name, gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);

		dpms = gt_hdmi->connector.dpms;
		conn_st = gt_hdmi->chip.hpd_status_detect(&gt_hdmi->chip);
		gt_hdmi->ac->update_eld(gt_hdmi->ac, gt_hdmi->connector.eld, MAX_ELD_BYTES);
		gt_hdmi->ac->report_jack(gt_hdmi->ac, dpms, conn_st);
	}
#endif

	if (encoder->crtc) {
		hdmi_info(gt_hdmi->dev,
			"gt %s encoder mode disable\n",gt_hdmi->name );
	} else {
		hdmi_info(gt_hdmi->dev,
			"gt %s encoder mode disable, crtc is null\n",gt_hdmi->name);
	}

	if (gt_hdmi->chip.encoder_disable)
		gt_hdmi->chip.encoder_disable(&gt_hdmi->chip);
}

static void gt_hdmi_encoder_mode_enable(struct drm_encoder *encoder)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(encoder, struct hdmi_device_t, encoder);

	if (NULL == encoder) {
		goto out;
	}

	if (encoder->crtc) {
		hdmi_info(gt_hdmi->dev,
			"gt %s encoder mode enable\n",gt_hdmi->name );
	} else {
		hdmi_info(gt_hdmi->dev,
			"gt %s encoder mode enable, crtc is null\n",gt_hdmi->name);
	}
	if (gt_hdmi->chip.encoder_enable)
		gt_hdmi->chip.encoder_enable(&gt_hdmi->chip, encoder->crtc);

	if (gt_hdmi->chip.output_mode &&
		is_output_type_dvi(gt_hdmi->chip.output_mode))
		goto out;

#ifdef CONFIG_DRM_GT_AUDIO
		if(gt_hdmi->ac && gt_hdmi->ac->report_jack) {
			int conn_st, dpms;
			hdmi_info(gt_hdmi->dev,
					"try report jack, conn_st:%d, has_audio:%d",
					gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);

			dpms = gt_hdmi->connector.dpms;
			conn_st = gt_hdmi->chip.hpd_status_detect(&gt_hdmi->chip);
			gt_hdmi->ac->update_eld(gt_hdmi->ac, gt_hdmi->connector.eld, MAX_ELD_BYTES);
			gt_hdmi->ac->report_jack(gt_hdmi->ac, dpms, conn_st);
		}
#endif

	hdmi_info(gt_hdmi->dev,  "enable to scramble work.\n");
	gt_hdmi->scramb_cnt = 0;
	gt_hdmi->scramb_gracecnt = 30;
	gt_hdmi->scramb_active = 1;
	queue_delayed_work(gt_hdmi->scrambwq, &gt_hdmi->scramble_work, msecs_to_jiffies(200));

out:
	atomic64_set(&gt_hdmi->modesetting, 0);
}

static int gt_hdmi_encoder_destroy(struct drm_encoder *encoder)
{
	BUG_ON(encoder);

	/*
	 * Called  drm_encoder_cleanup
	 */
	if (encoder->funcs && encoder->funcs->destroy)
		encoder->funcs->destroy(encoder);

	return 0;
}

static const struct drm_encoder_funcs s_gt_hdmi_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static const struct drm_encoder_helper_funcs s_gt_hdmi_encoder_helper_funcs = {
	//.mode_valid = gt_hdmi_encoder_mode_valid,
	.mode_fixup = gt_hdmi_encoder_mode_fixup,
	.mode_set = gt_hdmi_encoder_mode_set,
	.disable = gt_hdmi_encoder_mode_disable,
	.enable = gt_hdmi_encoder_mode_enable,
};

static int gt_hdmi_encoder_create(struct drm_device *drm_dev,
		struct drm_encoder *encoder, unsigned int possible_crtc)
{
	int retcode = 0;

	drm_encoder_helper_add(encoder, &s_gt_hdmi_encoder_helper_funcs);
	retcode = drm_encoder_init(drm_dev, encoder,
		&s_gt_hdmi_encoder_funcs, DRM_MODE_ENCODER_TMDS, NULL);

	encoder->possible_crtcs = possible_crtc;

	return retcode;
}

static void gt_hdmi_connector_destroy_func(struct drm_connector *connector)
{
	if (!connector) {
		return;
	}

	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static int gt_hdmi_get_edid(void *data, u8 * buf, u32 block, size_t len)
{
	struct hdmi_chip_t *chip = (struct hdmi_chip_t *)data;

	gt_memcpy(buf, chip->edid_buf + EDID_LENGTH * block, len);

	return 0;
}

static int gt_hdmi_ddc_status_detect(struct drm_connector *connector)
{
	struct hdmi_device_t *gt_hdmi = container_of(connector, struct hdmi_device_t, connector);
	u8 out = 0x0;
	u8 buf[8];
	int ret;
	struct i2c_msg msgs[] = {
		{
			.addr = DDC_ADDR,
			.flags = 0,
			.len = 1,
			.buf = &out,
		},
		{
			.addr = DDC_ADDR,
			.flags = I2C_M_RD,
			.len = 8,
			.buf = buf,
		}
	};

	ret = i2c_transfer(&gt_hdmi->i2c->adapter, msgs, 2);
	if (ret != 2)
		/* Couldn't find an accessible DDC on this connector */
		return false;

	/* Probe also for valid EDID header
	 * EDID header starts with:
	 * 0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00.
	 * Only the first 6 bytes must be valid as
	 * drm_edid_block_valid() can fix the last 2 bytes */
	if (gt_drm_edid_header_is_valid(buf) < 6) {
		/* Couldn't find an accessible EDID on this
		 * connector */
		return false;
	}
	return true;
}

static int gt_hdmi_adapter_dongle(struct hdmi_device_t *gt_hdmi, struct edid *edid)
{

	if (!gt_hdmi || !edid) {
		return -1;
	}

	switch (gt_hdmi->connector_type) {
	case DRM_MODE_CONNECTOR_HDMIA:
		/*if it is not hdmi sink for hdmi source, to force set dvi mode.*/
		gt_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_HDMIA;
		if (!drm_detect_hdmi_monitor(edid)) {
			gt_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_DVII;
		}
		break;

	case DRM_MODE_CONNECTOR_DVII:
		/*if it is hdmi sink for dvi source, to force set hdmi mode.*/
		gt_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_DVII;
		if (drm_detect_hdmi_monitor(edid)) {
			gt_hdmi->chip.output_mode->mode_connector_item = DRM_MODE_CONNECTOR_HDMIA;
		}
		break;

	default:
		break;
	}

	return 0;
}

static int gt_hdmi_get_modes_by_autoedid(struct drm_connector *connector)
{
	int nums = 0;
	struct edid *edid = NULL;
	struct hdmi_device_t *gt_hdmi = NULL;

	if (!connector)
		return nums;

	gt_hdmi = container_of(connector, struct hdmi_device_t, connector);

	if (gt_hdmi->chip.hdmi_edid_read) {
		gt_hdmi->chip.hdmi_edid_read(&gt_hdmi->chip);
		edid = drm_do_get_edid(connector, gt_hdmi_get_edid, &gt_hdmi->chip);
	} else if (gt_hdmi->chip.connector_get_edid) {
		edid = gt_hdmi->chip.connector_get_edid(connector, &gt_hdmi->chip);
	}

	if (edid) {
		gt_hdmi->chip.max_clk = gtdpu_conn_get_monitor_max_clk((u8 *)edid);
		if (gt_hdmi->chip.max_clk <= 0) {
			gt_hdmi->chip.max_clk = 605000;
			hdmi_info(gt_hdmi->dev,  "get hdmi max clock err, set it to 605MHz\n");
		}

		if (gt_hdmi->chip.replace_timing)
			gtdpu_detailed_block_monitor_range_replace((u8 *)edid);

		hdmi_info(gt_hdmi->dev,  "update_eld");
		nums = drm_add_edid_modes(connector, edid);

#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif
		gt_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		gt_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);

		gt_hdmi_adapter_dongle(gt_hdmi, edid);

		drm_connector_update_edid_property(connector, edid);
#ifdef CONFIG_DRM_GT_AUDIO
		if(gt_hdmi->ac){
			gt_hdmi->ac->has_audio = gt_hdmi->chip.sink_has_audio;
			hdmi_info(gt_hdmi->dev,
					"update audio param, conn_st:%d, has_audio:%d",
					gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);
		}
#endif
		kfree(edid);
	}

	return nums;
}

static int gt_hdmi_get_modes_by_strpush(struct drm_connector *connector)
{
	int ret = -1;
	int nums = 0;
	struct hdmi_device_t *gt_hdmi = NULL;
	unsigned char str_push_edid_buf[GTHDMI_EDID_BUF_LEN] = {0};

	if (!connector)
		return nums;

	gt_hdmi = container_of(connector, struct hdmi_device_t, connector);

	ret = hal_hdmi_edid_data(gt_hdmi->dev->parent, gt_hdmi->hdmi_id,
			str_push_edid_buf);
	if (ret < 0) {
		hdmi_info(gt_hdmi->dev,  \
				"hal_hdmi_edid_data failed, return %d.\n", ret);
		return nums;
	}

	nums = gtdpu_str_push_edid(str_push_edid_buf, connector);
	drm_mode_sort(&connector->probed_modes);

	return nums;
}

static int gt_hdmi_get_modes_by_userdefine(struct drm_connector *connector)
{
	int ret = -1;
	int nums = 0;
	struct edid *edid = NULL;
	struct hdmi_device_t *gt_hdmi = NULL;

	if (!connector)
		return nums;

	gt_hdmi = container_of(connector, struct hdmi_device_t, connector);

	ret = hal_hdmi_edid_data(gt_hdmi->dev->parent, gt_hdmi->hdmi_id,
			gt_hdmi->chip.edid_buf);
	if (ret < 0) {
		hdmi_info(gt_hdmi->dev,  \
				"hal_hdmi_edid_data failed, return %d.\n", ret);
		return nums;
	}

	edid = drm_do_get_edid(connector, gt_hdmi_get_edid, &gt_hdmi->chip);
	if (edid) {
		gt_hdmi->chip.max_clk = gtdpu_conn_get_monitor_max_clk((u8 *)edid);
		if (gt_hdmi->chip.max_clk <= 0) {
			gt_hdmi->chip.max_clk = 605000;
			gtdpu_err(gt_hdmi->dev, "get hdmi max clock err, set it to 605MHz\n");
		}
		hdmi_info(gt_hdmi->dev, "update eld");
		nums = drm_add_edid_modes(connector, edid);

#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif

		gt_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		gt_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);

		gt_hdmi_adapter_dongle(gt_hdmi, edid);

		drm_connector_update_edid_property(connector, edid);
#ifdef CONFIG_DRM_GT_AUDIO
		if(gt_hdmi->ac){
			gt_hdmi->ac->has_audio = gt_hdmi->chip.sink_has_audio;
			hdmi_info(gt_hdmi->dev,
					"update audio param, conn_st:%d, has_audio:%d",
					gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);
		}
#endif
		kfree(edid);
	}

	return nums;
}

static bool gt_hdmi_native_mode_filter(struct drm_connector *connector,
		struct drm_display_mode *mode, bool priv_filter)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	struct drm_encoder *encoder = &gt_hdmi->encoder;
#endif
	int ret = 0;

	// TBD: current not support 4k scaler
	if (mode && priv_filter && ((mode->hdisplay > GTDPU_COMBINE_WIDTH) ||
		(mode->vdisplay > GTDPU_COMBINE_HEIGHT)) )
		return true;

	if (mode && gt_hdmi->chip.max_clk > 0 && \
		mode->clock > gt_hdmi->chip.max_clk)
		return true;

	if (connector->helper_private && connector->helper_private->mode_valid) {
		ret = connector->helper_private->mode_valid(&gt_hdmi->connector, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}

#if ((DRM_VERSION >= KERNEL_VERSION(4, 13, 0)))
	if (encoder->helper_private && encoder->helper_private->mode_valid) {
		ret = encoder->helper_private->mode_valid(&gt_hdmi->encoder, mode);
		if (ret != MODE_OK) {
			return true;
		}
	}
#endif
	return false;
}

/**
 * gt_hdmi_connector_helper_get_modes - Get hdmi supports mode;
 *
 * Returns:
 * Number of supported modes
 *
 * 1. read edid and get modes
 * 2. push string
 * 3. replace timing
 * 4. get native mode
 * 5. add common modes
 */
static int gt_hdmi_connector_helper_get_modes(
		struct drm_connector *connector)
{
	int nums = 0;
	struct drm_display_mode *native_mode;
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	switch (gt_hdmi->chip.hal_edid_mode) {
		case EDID_AUTO_READ:
			nums = gt_hdmi_get_modes_by_autoedid(connector);
			break;
		case EDID_STR_PUSH:
			nums = gt_hdmi_get_modes_by_autoedid(connector);
			if (nums <= 0) {
				hdmi_info(gt_hdmi->dev,  "zoom:can't read autoedid.\n");
			}
			nums += gt_hdmi_get_modes_by_strpush(connector);
			if (nums <= 0) {
				hdmi_info(gt_hdmi->dev,  "zoom:can't parse strpush.\n");
			}
			break;
		case EDID_USER_DEFINE:
			nums = gt_hdmi_get_modes_by_userdefine(connector);
			break;
		default:
			nums = 0;
			gtdpu_warn(gt_hdmi->dev, "unkown hal_edid_mode(%d)!\n", gt_hdmi->chip.hal_edid_mode);
			break;
	}

	if (!nums) {
		gt_hdmi->chip.modes = gtdpu_add_modes_without_edid(connector, NULL);
	} else {
		gt_hdmi->chip.modes = nums;
	}

	// drop repeat mode
	gt_hdmi->chip.modes -= gtdpu_modes_drop_repeat(&gt_hdmi->connector);

	//sort modes
	drm_mode_sort(&connector->probed_modes);

	// fixup 1366x768 and more modes for /dev/fb0
	gtdpu_modes_fixup_preferred_nonaligned_modes(&gt_hdmi->connector);

	// get native mode
	native_mode = gtdpu_get_native_mode(connector, gt_hdmi_native_mode_filter);

	// add common mode
	if (native_mode) {
		hdmi_info(gt_hdmi->dev,  "%s native mode: "DRM_MODE_FMT ", status:%d\n",
			gt_hdmi->name, DRM_MODE_ARG(native_mode), native_mode->status);
		gt_memcpy(&gt_hdmi->native_mode, native_mode, sizeof(gt_hdmi->native_mode));
		drm_mode_set_crtcinfo(&gt_hdmi->native_mode, CRTC_INTERLACE_HALVE_V);
		gt_hdmi->native_mode.status = MODE_OK;
		gt_hdmi->chip.modes += gtdpu_connector_add_common_modes(connector, &gt_hdmi->native_mode, false);
	}

	return gt_hdmi->chip.modes;
}


/**
 * gt_hdmi_connector_detect_ctx - check hdmi hpd status
 *
 * Returns:
 * drm_connector_status. (connector_status_connected or connector_status_disconnected)
 */
static int gt_hdmi_connector_detect_ctx(struct drm_connector *connector,
		struct drm_modeset_acquire_ctx *ctx, bool force)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	int conn_status = connector_status_unknown;

	if (gt_hdmi->chip.connector_detect) {
		conn_status = gt_hdmi->chip.connector_detect(&gt_hdmi->chip);
	} else {
		gtdpu_warn(gt_hdmi->dev, "Not support %s chip detect handle !!!\n",
				gt_hdmi->name);
	}

	hdmi_info(gt_hdmi->dev,
		"%s connector detect ctx: %s\n", gt_hdmi->name,
		gt_drm_get_connector_status_name(conn_status));

	return conn_status;
}

static enum drm_connector_status gt_hdmi_detect(struct drm_connector *connector, bool force)
{
    return gt_hdmi_connector_detect_ctx(connector, NULL, force);
}

static enum drm_mode_status gt_hdmi_connector_helper_mode_valid(
		struct drm_connector *connector, struct drm_display_mode *mode)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	int status = MODE_OK;

	if (mode->flags & DRM_MODE_FLAG_DBLSCAN) {
		return MODE_NO_DBLESCAN;
	}

	if (mode->flags & DRM_MODE_FLAG_INTERLACE) {
		return MODE_NO_INTERLACE;
	}

	if (!s_hdmi_support_75hz && drm_mode_vrefresh(mode) == 75) {
		return MODE_NOMODE;
	}

	if (mode->clock > 600000) {
		return MODE_CLOCK_HIGH;
	}

//#define HDMI_MODE_NEED_BYPASS
#ifdef  HDMI_MODE_NEED_BYPASS
	{
		struct mfc_monitor_info monitor = MFC_MONITOR_INFO_ITEM("LEN", "*");
		struct resolution_info  resolution[] = {
			RESOLUTION_INFO_ITEM(1920, 1080, 75, -1),
			RESOLUTION_INFO_ITEM(2560, 1440, 60, -1),
		};
		struct resolution_info  *match_resolution = NULL;

		if (gtdpu_is_mfc_monitor_match((unsigned char *)gt_hdmi->chip.edid_buf, &monitor)) {
			match_resolution = gtdpu_resolution_match(mode, resolution, GT_ARRAY_SIZE(resolution));
			if (match_resolution) {
				hdmi_info(gt_hdmi->dev,  \
						"match resolution[%dx%d@%d %dKHz], drop it for monitor(LEN/*)\n",
						match_resolution->hdisplay, match_resolution->vdisplay, match_resolution->vrefresh);
				return MODE_NOMODE;
			}
		}
	}
#endif

	if (gt_hdmi->chip.connector_mode_valid) {
		status = gt_hdmi->chip.connector_mode_valid(&gt_hdmi->chip, mode);
		if (status != MODE_OK)
			return status;
	}

	return MODE_OK;
}

#if ((DRM_VERSION >= KERNEL_VERSION(4, 10, 0)))
static void gt_hdmi_connector_print_state(struct drm_printer *p,
					     const struct drm_connector_state *state)
{

}
#endif

static int gt_hdmi_connector_late_register(struct drm_connector *connector)
{
	struct hdmi_device_t * gt_hdmi = container_of(connector, struct hdmi_device_t, connector);
	int ret = 0;
	struct i2c_adapter *adapter = NULL;
	struct device *adapter_dev  = NULL;

	BUG_ON(!(connector->kdev));

	if (gt_hdmi->i2c) {
		adapter = &gt_hdmi->i2c->adapter;
		adapter_dev  = &adapter->dev;

		ret = sysfs_create_link(&connector->kdev->kobj,
				&adapter_dev->kobj,
				adapter_dev->kobj.name);
		if (ret) {
			gtdpu_err(gt_hdmi->dev, "%s sysfs link create failed-%d\n",
				  adapter_dev->kobj.name, ret);
		}
	}

#if defined(CONFIG_DEBUG_FS)
	ret = gt_hdmi_custom_debugfs_create(connector->debugfs_entry, &gt_hdmi->chip);
	if (ret) {
		gtdpu_err(gt_hdmi->dev, "%s custom debugfs create failed-%d\n",
			  gt_hdmi->name, ret);
	}
#endif

	hal_dev_enable_irq(gt_hdmi->dev->parent, gt_hdmi->chip.hal_module);

	queue_delayed_work(gt_hdmi->hpdwq, &gt_hdmi->hotplug_work, 0);

	return 0;
}

static void gt_hdmi_connector_early_unregister(struct drm_connector *connector)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	struct i2c_adapter *adapter = NULL;
	struct device *adapter_dev  = NULL;

	BUG_ON(!(connector->kdev));

	hal_dev_disable_irq(gt_hdmi->dev->parent, gt_hdmi->chip.hal_module);

#if defined(CONFIG_DEBUG_FS)
	gt_hdmi_custom_debugfs_remove(connector->debugfs_entry, &gt_hdmi->chip);
#endif

	if (gt_hdmi->i2c) {
		adapter = &gt_hdmi->i2c->adapter;
		adapter_dev  = &adapter->dev;
		sysfs_remove_link(&connector->kdev->kobj, adapter_dev->kobj.name);
	}
}

static struct drm_connector_helper_funcs s_gt_hdmi_connector_helper_funcs = {
	.get_modes  = gt_hdmi_connector_helper_get_modes,
	.mode_valid = gt_hdmi_connector_helper_mode_valid,
#if ((DRM_VERSION >= KERNEL_VERSION(4, 12, 0)))
	.detect_ctx = gt_hdmi_connector_detect_ctx,
#endif
};


static const struct drm_connector_funcs s_gt_hdmi_connector_funcs = {
	.dpms       = drm_helper_connector_dpms,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = gt_hdmi_connector_destroy_func,
#if ((DRM_VERSION >= KERNEL_VERSION(4, 10, 0)))
	.atomic_print_state = gt_hdmi_connector_print_state,
#endif
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
	.late_register = gt_hdmi_connector_late_register,
	.early_unregister = gt_hdmi_connector_early_unregister,
	.detect = gt_hdmi_detect,
};

static int gt_hdmi_connector_create(struct drm_device *drm_dev,
				struct drm_connector *connector)
{
	int retcode = 0;
	struct hdmi_device_t * hdmi_dev = container_of(connector, struct hdmi_device_t, connector);

	BUG_ON(!drm_dev);
	BUG_ON(!connector);

	drm_connector_helper_add(connector, &s_gt_hdmi_connector_helper_funcs);

	if (hdmi_dev->chip.output_mode) {
		hdmi_dev->connector_type = hdmi_dev->chip.output_mode->mode_connector_item;
	} else {
		hdmi_dev->connector_type = DRM_MODE_CONNECTOR_HDMIA;
	}

	retcode = drm_connector_init(drm_dev, connector, &s_gt_hdmi_connector_funcs,
			hdmi_dev->connector_type);

	connector->polled = DRM_CONNECTOR_POLL_HPD;
	connector->dpms = DRM_MODE_DPMS_OFF;
	connector->interlace_allowed = false;
	connector->doublescan_allowed = false;
	connector->display_info.subpixel_order = SubPixelHorizontalRGB;

	return retcode;
}

static int gt_hdmi_connector_destroy(struct drm_connector *connector)
{
	struct drm_mode_object *conn_obj = &connector->base;

	while (kref_read(&conn_obj->refcount) > 0)
		drm_connector_put(connector); // will be called drm_connector_free

	return 0;
}


static int gt_hdmi_connector_attach_encoder(
			struct hdmi_device_t *gt_hdmi)
{
	int retcode = 0;

	retcode = gt_hdmi_encoder_create(gt_hdmi->drm_dev, &gt_hdmi->encoder,
		gt_hdmi->chip.possible_crtc);
	if (retcode) {
		gtdpu_err(gt_hdmi->dev, "%s encoder create failed-%d\n",
				  gt_hdmi->name, retcode);
		return retcode;
	}

	retcode = gt_hdmi_connector_create(gt_hdmi->drm_dev, &gt_hdmi->connector);
	if (retcode) {
		gtdpu_err(gt_hdmi->dev, "%s connector create failed-%d\n",
			gt_hdmi->name, retcode);
		goto err_connector_create;
	}

	// only set connector->possible_encoder,
	retcode = drm_connector_attach_encoder(&gt_hdmi->connector, &gt_hdmi->encoder);
	if (retcode) {
		gtdpu_err(gt_hdmi->dev, "%s connector attach encoder failed-%d\n",
			gt_hdmi->name, retcode);
		goto error_attach_failed;
	}

	hdmi_info(gt_hdmi->dev,  "%s attach encoder-%d connector-%d\n",
		gt_hdmi->name, gt_hdmi->encoder.base.id, gt_hdmi->connector.base.id);

	return retcode;

error_attach_failed:
	gt_hdmi_connector_destroy(&gt_hdmi->connector);
err_connector_create:
	gt_hdmi_encoder_destroy(&gt_hdmi->encoder);

	return retcode;
}

static void gt_hdmi_cleanup_connector_encoder(
			struct hdmi_device_t *gt_hdmi)
{
	/*
	 * After the hdmi reference count is zeroed, drm_connector_free will be called
	 * encoder autorelease function without reference counting
	 */
	gt_hdmi_connector_destroy(&gt_hdmi->connector);
	gt_hdmi_encoder_destroy(&gt_hdmi->encoder);
}

static int gt_hdmi_edid_fixup(struct drm_connector *connector)
{
	int ret = -1;
	struct edid *edid = NULL;
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	gt_hdmi = container_of(connector, struct hdmi_device_t, connector);
	switch (gt_hdmi->chip.hal_edid_mode) {
	case EDID_AUTO_READ:
		if (gt_hdmi->chip.hdmi_edid_read) {
			gt_hdmi->chip.hdmi_edid_read(&gt_hdmi->chip);
			edid = drm_do_get_edid(connector, gt_hdmi_get_edid, &gt_hdmi->chip);
		} else if (gt_hdmi->chip.connector_get_edid) {
			edid = gt_hdmi->chip.connector_get_edid(connector, &gt_hdmi->chip);
		}
		break;
	case EDID_USER_DEFINE:
		ret = hal_hdmi_edid_data(gt_hdmi->dev->parent, gt_hdmi->hdmi_id,
				gt_hdmi->chip.edid_buf);
		if (ret < 0) {
			hdmi_info(gt_hdmi->dev,  \
					"hal_hdmi_edid_data failed, return %d.\n", ret);
			return -1;
		}
		edid = drm_do_get_edid(connector, gt_hdmi_get_edid, &gt_hdmi->chip);
		break;
	default:
		return -1;
	}

	if (edid) {
#if (DRM_VERSION <= KERNEL_VERSION(4, 16, 0))
		drm_edid_to_eld(connector, edid);
#endif

		gt_hdmi->chip.sink_is_hdmi = drm_detect_hdmi_monitor(edid);
		gt_hdmi->chip.sink_has_audio = drm_detect_monitor_audio(edid);

		gt_hdmi_adapter_dongle(gt_hdmi, edid);

		drm_connector_update_edid_property(connector, edid);

		kfree(edid);
		return 0;
	}

	return -1;
}

static void gt_hdmi_hpdin_fixup(struct drm_connector *connector)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);
	struct drm_encoder *encoder = &gt_hdmi->encoder;
	struct drm_crtc *crtc = encoder->crtc;

	if (crtc && crtc->state && crtc->state->active && !atomic64_read(&gt_hdmi->modesetting)) {
		struct drm_display_mode *mode = NULL;

		mode = gt_hdmi->chip.adjusted_mode;

		if (IS_ERR_OR_NULL(mode)) {
			hdmi_info(gt_hdmi->dev,  "recover-mode is NULL\n");
			return;
		}

		hdmi_info(gt_hdmi->dev,  "%s start recover-%dx%d@%d\n",
				gt_hdmi->name, mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode));

		/*
		 * TODO:
		 * if edid is null, we try to update it.
		 * but there is not a good way, we will find out why we didn't get
		 * the edid and solve it, rather than the gt_hdmi_edid_fixup function.
		 * */

		if (!((gt_drm_edid_block_checksum(gt_hdmi->chip.edid_buf) == 0) &&
			(gt_drm_edid_header_is_valid(gt_hdmi->chip.edid_buf) == 8))) {
			hdmi_info(gt_hdmi->dev,  "Try to fixup edid\n");
			gt_hdmi_edid_fixup(connector);
		}
		gt_hdmi_encoder_mode_set(encoder, NULL, mode);
#ifdef CONFIG_DRM_GT_AUDIO
		if(gt_hdmi->ac){
			gt_hdmi->ac->has_audio = gt_hdmi->chip.sink_has_audio;
			hdmi_info(gt_hdmi->dev,
							"conn_st:%d, has_audio:%d",
							gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);
		}
#endif
		gt_hdmi_encoder_mode_enable(encoder);
	}

	hdmi_set_hpg_status(&gt_hdmi->chip, 1);
}


static int gt_hdmi_hpdout_clean(struct drm_connector *connector)
{
	struct hdmi_device_t *gt_hdmi
		= container_of(connector, struct hdmi_device_t, connector);

	hdmi_set_hpg_status(&gt_hdmi->chip, 0);

	gt_hdmi->chip.sink_is_hdmi   = false;
	gt_hdmi->chip.sink_has_audio = false;
	hdmi_info(gt_hdmi->dev,  "edid clear ok\n");

#ifdef CONFIG_DRM_GT_AUDIO
	if(gt_hdmi->ac){
		gt_hdmi->ac->has_audio = 0;
		gtdpu_connector_clear_eld(&gt_hdmi->connector);
		hdmi_info(gt_hdmi->dev, "conn_st:%d, has_audio:%d",
						gt_hdmi->ac->conn_st, gt_hdmi->ac->has_audio);
	}
#endif
	drm_connector_update_edid_property(connector, NULL);
	gt_memset(&gt_hdmi->native_mode, 0, sizeof(gt_hdmi->native_mode));
	gt_memset(gt_hdmi->chip.edid_buf, 0, GTHDMI_EDID_BUF_LEN);

	drm_connector_update_edid_property(connector, NULL);

	gt_hdmi->chip.max_clk = 0;

	return 0;
}

/*
 * If the ddc detection is more than N times, the HPD signal is not elevated,
 * which means that the display should be pulled out.
 *
 */
#define GTHDMI_HPDOUT_THRESHOLD (5U)
static void __attribute__ ((unused)) gt_hdmi_hotplug_work(struct work_struct *work)
{
	struct hdmi_device_t *gt_hdmi =
		container_of(work, struct hdmi_device_t, hotplug_work.work);

	int hpd_status, ddc_status;

	/* when hpd-pin is high, should to report hpd-in. */
	if (gt_hdmi->chip.hpd_status_detect) {
		hpd_status = gt_hdmi->chip.hpd_status_detect(&gt_hdmi->chip);
		if (hpd_status) {
			gt_hdmi->hpdout_cnt = 0;
			gt_hdmi_hpdin_fixup(&gt_hdmi->connector);
			drm_helper_hpd_irq_event(gt_hdmi->drm_dev);
			return;
		}
	/* in order to support old way.*/
	} else if (gt_hdmi->chip.connector_detect) {
		if (connector_status_connected == \
			gt_hdmi->chip.connector_detect(&gt_hdmi->chip)) {
			gt_hdmi->hpdout_cnt = 0;
			drm_helper_hpd_irq_event(gt_hdmi->drm_dev);
			return;
		}
	}

	/* when hpd low happend, should to read ddc status,
	 * when ddc is okay, to ignore it.
	 * because it could be connected, the user to close hdmi signel out.
	 * when ddc is failed, to report hpd-out.
	 * */
	if (!is_flags_skip_ddc_detect(gt_hdmi->chip.output_mode)) {
		ddc_status = gt_hdmi_ddc_status_detect(&gt_hdmi->connector);
		hdmi_info(gt_hdmi->dev,  \
				"loop ddc_detect:%s, hpdout_cnt:%d.\n",\
				ddc_status ? "okay" : "failed", gt_hdmi->hpdout_cnt);
		if (ddc_status && gt_hdmi->hpdout_cnt < GTHDMI_HPDOUT_THRESHOLD) {
			gt_hdmi->hpdout_cnt ++;
			cancel_delayed_work(&gt_hdmi->hotplug_work);
			queue_delayed_work(gt_hdmi->hpdwq, &gt_hdmi->hotplug_work, \
					msecs_to_jiffies(1000));
			return;
		} else {
			gt_hdmi->hpdout_cnt = 0;
			gt_hdmi_hpdout_clean(&gt_hdmi->connector);
			drm_helper_hpd_irq_event(gt_hdmi->drm_dev);
			return;
		}
	} else {
		//force to report hpd-out
		gt_hdmi_hpdout_clean(&gt_hdmi->connector);
		drm_helper_hpd_irq_event(gt_hdmi->drm_dev);
		return;
	}
}

static int __attribute__ ((unused)) gt_hdmi_read_channel_cnterr(
		struct hdmi_device_t *gt_hdmi, char *str, int len)
{
	int n = 0, i = 0;
	unsigned char b[8] = {0}, sum = 0;
	unsigned int  cnt[3] = {0};

	gt_drm_scdc_read(&gt_hdmi->i2c->adapter, 0x50, b, 7);

	for (i=0; i<7; i++) {
		sum += b[i];
	}

	cnt[0] = b[0] | (b[1]<<8);
	cnt[1] = b[2] | (b[3]<<8);
	cnt[2] = b[4] | (b[5]<<8);

	n += snprintf(str + n, len - n, "cnt[0]:%8d,", cnt[0] & 0x7FFFU);
	n += snprintf(str + n, len - n, "cnt[1]:%8d,", cnt[1] & 0x7FFFU);
	n += snprintf(str + n, len - n, "cnt[2]:%8d,", cnt[2] & 0x7FFFU);

	if (sum) {
		n += snprintf(str + n, len - n, "  CheckSumFailed[0x%02x]", sum);
	}

	//hdmi_info(gt_hdmi->dev,  "ErrCnt: %s\n", str);

	return n;
}

static void __attribute__ ((unused)) gt_hdmi_scramble_work(struct work_struct *work)
{
	struct hdmi_device_t *gt_hdmi =
		container_of(work, struct hdmi_device_t, scramble_work.work);
	int ret = 0;
	int keep_on = 0;
	char string[64] = {0};

	if (!gt_hdmi->scramb_active) {
		return ;
	}

#if 0
	/*Only debug for SCDC scramble to show channel error status.*/
	gt_hdmi_read_channel_cnterr(gt_hdmi, string, sizeof(string));
#endif

	if (gt_hdmi->chip.try_scramble) {
		ret = gt_hdmi->chip.try_scramble(&gt_hdmi->chip, 0x3);
	}

	if (ret > 0) {
		hdmi_info(gt_hdmi->dev,  "unscramble OK(cnt:%d). %s\n", gt_hdmi->scramb_cnt, string);
	} else if (ret < 0) {
		if (gt_hdmi->scramb_cnt < 30) {
			gt_hdmi->scramb_cnt++;
			keep_on = 1;
			if (!(gt_hdmi->scramb_cnt % 5))
				hdmi_info(gt_hdmi->dev,  "scramble again(cnt:%d). %s\n", gt_hdmi->scramb_cnt, string);
		} else {
			hdmi_info(gt_hdmi->dev,  "scramble Failed(cnt:%d). %s\n", gt_hdmi->scramb_cnt, string);
		}
	} else {
		if (gt_hdmi->scramb_gracecnt) {
			gt_hdmi->scramb_gracecnt--;
			keep_on = 1;
			if (!(gt_hdmi->scramb_gracecnt % 5))
				hdmi_info(gt_hdmi->dev,  "scramble grace(cnt:%d). %s\n", gt_hdmi->scramb_gracecnt, string);
		} else {
			hdmi_info(gt_hdmi->dev,  "scramble OK(cnt:%d). %s\n", gt_hdmi->scramb_cnt, string);
		}
	}

	if (keep_on)
		queue_delayed_work(gt_hdmi->scrambwq, &gt_hdmi->scramble_work, msecs_to_jiffies(200));

	return ;
}

static void gt_hdmi_irq_handle(void *data)
{
	unsigned int irq_event = 0;
	struct hdmi_device_t *gt_hdmi = (struct hdmi_device_t *)data;
	unsigned int delay_ms   = 10;
	int hpd_status = 0;

	BUG_ON(!gt_hdmi);

	if (gt_hdmi->chip.irq_handle) {
		irq_event = gt_hdmi->chip.irq_handle(&gt_hdmi->chip);
	} else {
		gtdpu_warn(gt_hdmi->dev, "Not support %s chip irq handle !!!\n",
				  gt_hdmi->name);
		return;
	}

	if (irq_event & HDMI_IRQ_HPD_MASK) {
		hpd_status = gt_hdmi->chip.hpd_status_detect(&gt_hdmi->chip);

#ifdef CONFIG_DRM_GT_AUDIO
		if (gt_hdmi->ac) {
			gt_hdmi->ac->conn_st = hpd_status;
		}
#endif

		if (!hpd_status) {
			/*
			 * when hpd is low, we need wait longer to skip abormal hpd signel.
			 * and if not specified, the default is 500ms.
			 * */
			if (is_flags_wait_longer_hpd(gt_hdmi->chip.output_mode)) {
				delay_ms = gt_hdmi->chip.output_mode->data;
			} else {
				delay_ms = 500;
			}
		} else {
			/*
			 * when hpd is high,
			 * longer delay to wait dvi sink ready, or will fail to read edid.
			 */
			if (gt_hdmi->chip.output_mode && is_output_type_dvi(gt_hdmi->chip.output_mode)) {
				delay_ms = 500;
			}
		}

		cancel_delayed_work(&gt_hdmi->hotplug_work);
		queue_delayed_work(gt_hdmi->hpdwq, &gt_hdmi->hotplug_work, msecs_to_jiffies(delay_ms));
		gt_hdmi->scramb_active = 0;
	}

	return;
}

void gt_hdmi_hw_irq_enable(struct hdmi_device_t *gt_hdmi, unsigned int flags)
{
	BUG_ON(!gt_hdmi);
	if (gt_hdmi->chip.irq_enable)
		return gt_hdmi->chip.irq_enable(&gt_hdmi->chip, flags);
}

void gt_hdmi_hw_irq_disable(struct hdmi_device_t *gt_hdmi, unsigned int flags)
{
	BUG_ON(!gt_hdmi);
	if (gt_hdmi->chip.irq_disable)
		return gt_hdmi->chip.irq_disable(&gt_hdmi->chip, flags);
}

static int gthdmi_get_clock(void *data)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->getscl) {
		return chip->chipi2c->getscl(chip);
	}

	return -1;
}

static int gthdmi_get_data(void *data)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->getsda) {
		return chip->chipi2c->getsda(chip);
	}

	return -1;
}

static void gthdmi_set_clock(void *data, int state)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->setscl) {
		chip->chipi2c->setscl(chip, state);
	}

	return;
}

static void gthdmi_set_data(void *data, int state)
{
	struct hdmi_device_i2c_t *i2c = (struct hdmi_device_i2c_t *)data;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->setsda) {
		chip->chipi2c->setsda(chip, state);
	}

	return;
}


#define I2C_RISEFALL_TIME 20
static int gthdmi_pre_xfer(struct i2c_adapter *adapter)
{
	int ret;
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->pre_xfer) {
		ret = chip->chipi2c->pre_xfer(chip);
		if (ret)
			return ret;
	}

	gthdmi_set_data(i2c, 1);
	gthdmi_set_clock(i2c, 1);

	gt_udelay(I2C_RISEFALL_TIME);
	return 0;
}

static void gthdmi_post_xfer(struct i2c_adapter *adapter)
{
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;

	if (chip && chip->chipi2c && chip->chipi2c->post_xfer) {
		chip->chipi2c->post_xfer(chip);
	}

	gthdmi_set_data(i2c, 1);
	gthdmi_set_clock(i2c, 1);

	gt_udelay(I2C_RISEFALL_TIME);
}

static int gthdmi_i2c_bit_xfer(struct i2c_adapter *adapter,
		struct i2c_msg *msgs, int num)
{
	extern const struct i2c_algorithm i2c_bit_algo;
	return i2c_bit_algo.master_xfer(adapter, msgs, num);
}

#define I2C_M_RD 0x0001
static int gthdmi_i2c_hw_xfer(struct i2c_adapter *adapter, struct i2c_msg *msgs, int num)
{
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&i2c->hdmi_dev->chip;
	int i, ret = -1;
	u16 addr, len;
	u8 *buf;

	if (chip && chip->chipi2c && chip->chipi2c->set_ddc_clk) {
		chip->chipi2c->set_ddc_clk(chip);
	}

	if (!chip || !chip->dev)
		return ret;

	for (i = 0; i < num; i++) {
		addr = (msgs[i].flags & I2C_M_RD) ? ((msgs[i].addr<<1) | 0x1) : (msgs[i].addr<<1);
		buf = msgs[i].buf;
		len = msgs[i].len;

		if (chip && chip->chipi2c && chip->chipi2c->hwi2c_transfer) {
			ret = chip->chipi2c->hwi2c_transfer(chip, addr, buf, len);
			if (ret < 0)
				break;
		}
	}

	if (!ret) {
		ret = num;
	}

	return ret;
}

static int gthdmi_i2c_xfer(struct i2c_adapter *adapter,
		struct i2c_msg *msgs, int num)
{
	int ret = -1;
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);
	struct hdmi_device_t *hdmi_dev = i2c->hdmi_dev;
	struct hdmi_chip_t   *chip = (struct hdmi_chip_t *)&hdmi_dev->chip;
	int (*xfer)(struct i2c_adapter *, struct i2c_msg *, int) = NULL;
	chip_type_e plat;

	plat = hal_get_chiptype(hdmi_dev->dev->parent);

	if (s_hdmi_hwi2c && (plat == CHIP_GT6908_SOC)) {
		xfer = gthdmi_i2c_hw_xfer;
	} else {
		xfer = gthdmi_i2c_bit_xfer;
	}

	gt_mutex_lock(chip->chipi2c->mutex);
	ret = xfer(adapter, msgs, num);
	gt_mutex_unlock(chip->chipi2c->mutex);

	return ret;
}

static u32 gthdmi_i2c_functionality(struct i2c_adapter *adapter)
{
	/*
	struct hdmi_device_i2c_t *i2c = container_of(adapter,
					       struct hdmi_device_i2c_t,
					       adapter);

	return i2c->algorithm->functionality(adap, msgs, num);
	*/

	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

static const struct i2c_algorithm gthdmi_algorithm = {
	.master_xfer = gthdmi_i2c_xfer,
	.functionality = gthdmi_i2c_functionality,
};

int gthdmi_i2c_init(struct hdmi_device_t *hdmi_dev)
{
	struct hdmi_device_i2c_t *i2c = NULL;
	struct i2c_adapter *adapter = NULL;
	struct i2c_algo_bit_data *bit_algo = NULL;

	i2c = kzalloc(sizeof(struct hdmi_device_i2c_t), gt_gfp_kernel);
	if (!i2c)
		return -ENOMEM;

	adapter = &i2c->adapter;
	bit_algo = &i2c->bit_algo;

	snprintf(adapter->name, sizeof(adapter->name), \
			"hdmi%d-i2c", hdmi_dev->hdmi_id);
	adapter->algo = &gthdmi_algorithm;
	adapter->class = I2C_CLASS_DDC;
	adapter->dev.parent = hdmi_dev->dev;
	adapter->nr = -1;
	adapter->owner = THIS_MODULE;

	bit_algo->getscl = gthdmi_get_clock;
	bit_algo->getsda = gthdmi_get_data;
	bit_algo->setscl = gthdmi_set_clock;
	bit_algo->setsda = gthdmi_set_data;
	bit_algo->pre_xfer = gthdmi_pre_xfer;
	bit_algo->post_xfer = gthdmi_post_xfer;
	bit_algo->udelay = I2C_RISEFALL_TIME;
	bit_algo->timeout = usecs_to_jiffies(2200);
	bit_algo->data = i2c;
	//bit_algo->can_do_atomic = true;
	adapter->retries = 3;
	adapter->algo_data = bit_algo;

	if (i2c_add_adapter(adapter)) {
		goto err;
	}

	//i2c->hw_ddc_ops  = ddc_ops;
	i2c->hdmi_dev = hdmi_dev;
	hdmi_dev->i2c = i2c;

	return 0;

err:
	kfree(i2c);
	return -EINVAL;
}


static int gt_hdmi_hw_init(struct hdmi_device_t *gt_hdmi)
{
	int retcode = 0;

	if (gt_hdmi->chip.hw_init)
		return gt_hdmi->chip.hw_init(&gt_hdmi->chip);

	return retcode;
}

static void gt_hdmi_hw_fini(struct hdmi_device_t *gt_hdmi)
{
	if (gt_hdmi->chip.hw_fini)
		gt_hdmi->chip.hw_fini(&gt_hdmi->chip);
}

static int gtdpu_hdmi_zoomenable_id(int hdmi_id)
{
	int zoom_id = -1;

	switch (hdmi_id) {
		case 0:
			zoom_id = HDMI0_ZOOM_ENABLE;
			break;
		case 1:
			zoom_id = HDMI1_ZOOM_ENABLE;
			break;
		case 2:
			zoom_id = HDMI2_ZOOM_ENABLE;
			break;
		case 3:
			zoom_id = HDMI3_ZOOM_ENABLE;
			break;
		default:
			break;
	}
	return zoom_id;
}

static int gt_hdmi_chip_init(struct hdmi_device_t *gt_hdmi)
{
	chip_type_e plat;
	int retcode = 0;
	int zoom_id = 0;

	zoom_id = gtdpu_hdmi_zoomenable_id(gt_hdmi->hdmi_id);

	if ((zoom_id >= 0) && (hal_zoom_is_enable(gt_hdmi->dev->parent, zoom_id) == 1)) {
		gt_hdmi->chip.hal_zoom_enable = false;
		gt_hdmi->zoom_id = zoom_id;
	} else {
		gt_hdmi->chip.hal_zoom_enable = false;
		gt_hdmi->zoom_id = 0;
	}

	gt_hdmi->chip.hal_edid_mode = hal_hdmi_edid_mode(gt_hdmi->dev->parent, gt_hdmi->hdmi_id);
	if (gt_hdmi->chip.hal_edid_mode < 0) {
		hdmi_info(gt_hdmi->dev, "hdmi-%d hal_hdmi_edid_mode failed, return %d\n",
			gt_hdmi->hdmi_id, gt_hdmi->chip.hal_edid_mode);
		gt_hdmi->chip.hal_edid_mode = EDID_AUTO_READ;
	}

	if (gt_hdmi->chip.hal_zoom_enable && \
			gt_hdmi->chip.hal_edid_mode == EDID_STR_PUSH) {
			gt_hdmi->chip.hal_edid_mode = EDID_ZOOM_ENABLE;
	}

	hdmi_info(gt_hdmi->dev, "hdmi-%d use edid mode:%d\n",
			gt_hdmi->hdmi_id, gt_hdmi->chip.hal_edid_mode);


	plat = hal_get_chiptype(gt_hdmi->dev->parent);
	switch(plat) {
	case CHIP_GT6912_SOC:
		hdmi_info(gt_hdmi->dev, "%s start init gt6912 soc.\n", gt_hdmi->name);
		gt_hdmi->chip.hal_zoom_enable = false;
		retcode = gt6912_soc_hdmi_chip_init(&gt_hdmi->chip,
				              gt_hdmi->dev, gt_hdmi->hdmi_id);
		break;
	case CHIP_GT6910_SOC:
		hdmi_info(gt_hdmi->dev, "%s start init gt6910 soc.\n", gt_hdmi->name);
		retcode = gt6910_soc_hdmi_chip_init(&gt_hdmi->chip,
				              gt_hdmi->dev, gt_hdmi->hdmi_id);
		break;
	case CHIP_GT6913_SOC:
		hdmi_info(gt_hdmi->dev, "%s start init gt6913 soc.\n", gt_hdmi->name);
		gt_hdmi->chip.hal_zoom_enable = false;
		retcode = gt6913_soc_hdmi_chip_init(&gt_hdmi->chip,
				              gt_hdmi->dev, gt_hdmi->hdmi_id);
		break;
	case CHIP_GT6908_SOC:
		hdmi_info(gt_hdmi->dev, "%s start init gt6908 soc.\n", gt_hdmi->name);
		retcode = gt6908_soc_hdmi_chip_init(&gt_hdmi->chip,
				              gt_hdmi->dev, gt_hdmi->hdmi_id);
		break;
	default:
		gtdpu_err(gt_hdmi->dev, "%s does not currently support %d platform.\n",
			gt_hdmi->name, plat);
		retcode = -EINVAL;
		break;
	}

	gthdmi_i2c_init(gt_hdmi);

	gt_hdmi->chip.output_mode = gtdpu_get_connector_output_mode(gt_hdmi->dev, gt_hdmi->chip.reg_module);
	if (gt_hdmi->chip.output_mode) {
		if (is_output_type_dvi(gt_hdmi->chip.output_mode) ||
			is_output_type_vga(gt_hdmi->chip.output_mode)  ||
			is_output_type_lvds(gt_hdmi->chip.output_mode)) {
			gt_hdmi->chip.max_width = 2048;
			gt_hdmi->chip.max_height = 2048;
		}
		hdmi_info(gt_hdmi->dev, "mode_connector:%d convert:%s",
			gt_hdmi->chip.output_mode->mode_connector_item, gt_hdmi->chip.output_mode->convert_name);
	}

	{
		struct hw_board_info board[] = {
			HW_BOARD_INFO_ITEM("ZY", "YF27_1"), //VGA+HDMI2VGA+DP2LVDS
			HW_BOARD_INFO_ITEM("ZY", "YF27_2"), //VGA+HDMI2VGA+DP2VGA
		};
		if (gtdpu_odm_pcb_match(gt_hdmi->dev->parent, board, GT_ARRAY_SIZE(board))) {
			gt_hdmi->chip.max_width = 1920;
			gt_hdmi->chip.max_height = 1080;
		}
	}

	hdmi_info(gt_hdmi->dev, "hdmi-%d zoom:%s",
			gt_hdmi->hdmi_id, gt_hdmi->chip.hal_zoom_enable ? "enable" : "disable");

	return retcode;
}

static void gt_hdmi_chip_fini(struct hdmi_device_t *gt_hdmi)
{
	chip_type_e plat;

	plat = hal_get_chiptype(gt_hdmi->dev->parent);
	switch(plat) {
	case CHIP_GT6912_SOC:
		gt6912_soc_hdmi_chip_fini(&gt_hdmi->chip);
		break;
	case CHIP_GT6910_SOC:
		gt6910_soc_hdmi_chip_fini(&gt_hdmi->chip);
		break;
	case CHIP_GT6913_SOC:
		gt6913_soc_hdmi_chip_fini(&gt_hdmi->chip);
		break;
	case CHIP_GT6908_SOC:
		gt6908_soc_hdmi_chip_fini(&gt_hdmi->chip);
		break;
	default:
		gtdpu_err(gt_hdmi->dev, "%s does not currently support %d platform.\n",
			gt_hdmi->name, plat);
		break;
	}
	return;
}

/**
 * gt_hdmi_bind - gengtu hdmi-driver initialization function
 * @dev: hdmi_device_t_info allocated when gt_hdmi_device_t_register
 * parent(dev->parent) is &pci_dev.dev
 * @master: component master
 * @data: point of struct drm_device
 *
 * This function initializes the gengtu hdmi device
 * 1. alloc and init hdmi handle
 * 2. hdmi hardware init
 * 3. hdmi ddcci supports
 * 4. creation and bonding of HDMI connectors and encoders
 * 5. setup hdmi irq
 *
 * Returns:
 * 0 if it is OK, errno otherwise.
 */

#ifdef CONFIG_DRM_GT_AUDIO
static void gt_hdmi_audio_enable(struct audio_conn *ac)
{
	struct hdmi_device_t *gt_hdmi = NULL;

	if (!ac || !ac->priv) {
		return ;
	}

	gt_hdmi = (struct hdmi_device_t *)ac->priv;
	if (gt_hdmi && gt_hdmi->chip.hdmi_enable_audio) {
		gt_hdmi->chip.hdmi_enable_audio(&gt_hdmi->chip);
	}
}

static void gt_hdmi_audio_disable(struct audio_conn *ac)
{
	struct hdmi_device_t *gt_hdmi = NULL;

	if (!ac || !ac->priv) {
		return ;
	}

	gt_hdmi = (struct hdmi_device_t *)ac->priv;
	if (gt_hdmi && gt_hdmi->chip.hdmi_disable_audio) {
		gt_hdmi->chip.hdmi_disable_audio(&gt_hdmi->chip);
	}
}
#endif

static int gt_hdmi_bind(struct device *dev,
					  struct device *master, void *data)
{
	int retcode = 0;
	struct drm_device * drm_dev = data;
	struct hdmi_device_t *gt_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);
#ifdef CONFIG_DRM_GT_AUDIO
	struct audio_conn *pac;
#endif
	struct drm_display_mode *gt_mode = NULL;

	hdmi_info(dev, "start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_HDMI0 + pdata->dev_idx)) {
		hdmi_info(dev,  "possible_crtc = 0, do not bind hdmi-%d\n", pdata->dev_idx);
		return retcode;
	}

	gt_hdmi = devm_kmalloc(dev, sizeof(*gt_hdmi), gt_gfp_kernel);
	if (!gt_hdmi) {
		gtdpu_err(dev, "Alloc hdmi-%d handle failed. Short of memory.\n",
			pdata->dev_idx);
		return -ENOMEM;
	}

	memset(gt_hdmi, 0, sizeof(*gt_hdmi));
	gt_hdmi->hdmi_id = pdata->dev_idx;
	gt_hdmi->dev = gt_get_device(dev);
	gt_hdmi->drm_dev = drm_dev;
	atomic64_set(&gt_hdmi->modesetting, 0);

	gt_hdmi->name = gt_kasprintf(gt_gfp_kernel, "gt-hdmi-%d", gt_hdmi->hdmi_id);
	if (!gt_hdmi->name) {
		gtdpu_err(dev, "Alloc hdmi-%d name failed. Short of memory.\n",
			pdata->dev_idx);
		retcode = -ENOMEM;
		goto err_out_name;
	}

	dev_set_drvdata(dev, gt_hdmi);
	hdmi_info(dev,  "%s start bind.\n", gt_hdmi->name);

	retcode = gt_hdmi_chip_init(gt_hdmi);
	if (retcode) {
		gtdpu_err(dev, "%s chip Init failed-%d.\n", gt_hdmi->name, retcode);
		goto err_chip_init;
	}

	retcode = gt_hdmi_hw_init(gt_hdmi);
	if (retcode) {
		gtdpu_err(dev, "%s hw Init failed-%d.\n", gt_hdmi->name, retcode);
		goto err_hw_init;
	}

	gt_mode = devm_kmalloc(dev, sizeof(*gt_mode), gt_gfp_kernel);
	if (!gt_mode) {
		gtdpu_err(dev, "Alloc hdmi-%d mode failed. Short of memory.\n",
			pdata->dev_idx);
		return -ENOMEM;
	}
	memset(gt_mode, 0, sizeof(*gt_mode));
	gt_hdmi->chip.adjusted_mode = gt_mode;

	retcode = gt_hdmi_connector_attach_encoder(gt_hdmi);
	if (retcode) {
		gtdpu_err(dev, "%s connector and encoder Init failed-%d.\n",
			gt_hdmi->name, retcode);
		goto err_attach_encoder;
	}

	retcode = hal_set_irq_handler(dev->parent,
		gt_hdmi->chip.hal_module, gt_hdmi_irq_handle, gt_hdmi);
	if (retcode) {
		gtdpu_err(dev, "%s irq Init failed-%d.\n", gt_hdmi->name, retcode);
		goto err_enable_irq;
	}

	gt_hdmi->hpdwq = create_singlethread_workqueue("hotplug");
	INIT_DELAYED_WORK(&gt_hdmi->hotplug_work, gt_hdmi_hotplug_work);

	gt_hdmi->scrambwq = create_singlethread_workqueue("scramble");
	INIT_DELAYED_WORK(&gt_hdmi->scramble_work, gt_hdmi_scramble_work);

#ifdef CONFIG_DRM_GT_AUDIO
	if (gt_hdmi->chip.output_mode && !is_output_type_dvi(gt_hdmi->chip.output_mode)) {
		pac = kzalloc(sizeof(struct audio_conn), gt_gfp_kernel);
		if (!pac) {
			gtdpu_err(dev, "Alloc audio_conn failed.\n");
			goto err_enable_irq;
		}
		pac->conn_st = 0; //todo
		pac->has_audio = 0; //todo
		pac->dev = (void *)gt_hdmi->dev;
		pac->id = gt_hdmi->hdmi_id;
		if(is_output_type_dp(gt_hdmi->chip.output_mode)){
			pac->type = GTAUDIO_CONNECTOR_TYPE_HDMI2DP;
		}else{
			pac->type = GTAUDIO_CONNECTOR_TYPE_HDMI;
		}
		pac->priv = (void *)gt_hdmi;
		pac->enable  = gt_hdmi_audio_enable;
		pac->disable = gt_hdmi_audio_disable;

		if (gtaudio_register_connector(pac)) {
			hdmi_info(dev,  "register audio_conn failed.\n");
			gt_hdmi->ac = NULL;
			kfree(pac);
		} else {
			hdmi_info(dev,  "register audio_conn ok.\n");
			gt_hdmi->ac = pac;
		}
	}
#endif

	hdmi_info(dev, "end\n");
	return retcode;

err_enable_irq:
	gt_hdmi_cleanup_connector_encoder(gt_hdmi);
err_attach_encoder:
	gt_hdmi_hw_fini(gt_hdmi);
err_hw_init:
	gt_hdmi_chip_fini(gt_hdmi);
err_chip_init:
	if (gt_hdmi->name)
		kfree(gt_hdmi->name);
err_out_name:
	put_device(dev);
	devm_kfree(dev, gt_hdmi);

	return retcode;
}

/**
* gt_hdmi_unbind - gengtu hdmi-driver initialization function
* @dev: hdmi_device_t_info allocated when gt_hdmi_device_t_register
* parent(dev->parent) is &pci_dev.dev
* @master: component master
* @data: point of struct drm_device
*
* This function deinitializes the gengtu hdmi device
*
* Once drm_dev_register is called, the connector and encoder
* resources are released by the drm_mode_config_cleanup function!!!
* So we don't need to worry about the release of the connector and encoder here.
*/
static void gt_hdmi_unbind(struct device *dev,
						  struct device *master, void *data)

{
	struct hdmi_device_t *gt_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);

	hdmi_info(dev, "start\n");
	BUG_ON(!dev);
	BUG_ON(!data);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_HDMI0 + pdata->dev_idx)) {
		hdmi_info(dev,  "possible_crtc = 0, do not unbind hdmi-%d\n", pdata->dev_idx);
		return;
	}

	gt_hdmi = dev_get_drvdata(dev);
	if (!gt_hdmi) {
		gtdpu_err(dev, "hdmi handle is NULL\n");
		return;
	}

	hal_set_irq_handler(dev->parent,
		gt_hdmi->chip.hal_module, NULL, NULL);

#ifdef CONFIG_DRM_GT_AUDIO
	if (gt_hdmi->ac) {
		gtaudio_unregister_connector(gt_hdmi->ac);
		kfree(gt_hdmi->ac);
		gt_hdmi->ac = NULL;
	}
#endif

	cancel_delayed_work_sync(&gt_hdmi->hotplug_work);
	cancel_delayed_work_sync(&gt_hdmi->scramble_work);

	gt_hdmi_hw_fini(gt_hdmi);
	gt_hdmi_chip_fini(gt_hdmi);
	kfree(gt_hdmi->name);

	put_device(dev);
	devm_kfree(dev, gt_hdmi);
	hdmi_info(dev, "end\n");
}


static const struct component_ops s_gt_hdmi_ops = {
	.bind = gt_hdmi_bind,
	.unbind = gt_hdmi_unbind,
};

static int gt_hdmi_probe(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	return component_add(&pdev->dev, &s_gt_hdmi_ops);
}

static int gt_hdmi_remove(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	component_del(&pdev->dev, &s_gt_hdmi_ops);
	return 0;
}

static struct platform_device_id s_gt_hdmi_device_id_table[] = {
	{.name = GT_HDMI_DEVICE_NAME, .driver_data = 0},
	{},
};
MODULE_DEVICE_TABLE(platform, s_gt_hdmi_device_id_table);

static int gt_hdmi_suspend(struct device *dev)
{
	struct hdmi_device_t *gt_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);

	hdmi_info(dev, "start\n");
	BUG_ON(!dev);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_HDMI0 + pdata->dev_idx)) {
		hdmi_info(dev,  "possible_crtc = 0, do not suspend hdmi-%d\n", pdata->dev_idx);
		return 0;
	}

	gt_hdmi = dev_get_drvdata(dev);
	if (!gt_hdmi) {
		gtdpu_err(dev, "hdmi handle is NULL\n");
		return 0;
	}

	hal_dev_disable_irq(dev->parent, gt_hdmi->chip.hal_module);

	cancel_delayed_work_sync(&gt_hdmi->hotplug_work);
	cancel_delayed_work_sync(&gt_hdmi->scramble_work);

	hdmi_info(dev, "end\n");
	return 0;
}

static int gt_hdmi_resume(struct device *dev)
{
	struct hdmi_device_t *gt_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(dev);

	hdmi_info(dev, "start\n");
	BUG_ON(!dev);

	if (!gtdpu_detect_is_valid_output(dev, CONNECTOR_M_HDMI0 + pdata->dev_idx)) {
		hdmi_info(dev,  "possible_crtc = 0, do not resume hdmi-%d\n", pdata->dev_idx);
		return 0;
	}

	gt_hdmi = dev_get_drvdata(dev);
	if (!gt_hdmi) {
		gtdpu_err(dev, "hdmi handle is NULL\n");
		return 0;
	}

	gt_hdmi_hw_init(gt_hdmi);

	hal_dev_enable_irq(dev->parent, gt_hdmi->chip.hal_module);

	queue_delayed_work(gt_hdmi->hpdwq, &gt_hdmi->hotplug_work, 0);

	hdmi_info(dev, "end\n");

	return 0;
}

static void gt_hdmi_shutdown(struct platform_device *pdev)
{
	struct hdmi_device_t *gt_hdmi = NULL;
	plat_data_t *pdata =  dev_get_platdata(&pdev->dev);

	if (!gtdpu_detect_is_valid_output(&pdev->dev, CONNECTOR_M_HDMI0 + pdata->dev_idx)) {
		hdmi_info(&pdev->dev, "possible_crtc = 0, do not shutdown hdmi-%d\n", pdata->dev_idx);
		return;
	}

	gt_hdmi =  gt_platform_get_drvdata(pdev);
	if (!gt_hdmi) {
		gtdpu_err(&pdev->dev, "hdmi handle is NULL\n");
		return;
	}

	gt_hdmi_encoder_mode_disable(&gt_hdmi->encoder);
	gt_hdmi_suspend(&pdev->dev);
}

static const struct dev_pm_ops gt_hdmi_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(gt_hdmi_suspend, gt_hdmi_resume)
};

struct platform_driver g_gtgpu_hdmi_driver = {
	.probe = gt_hdmi_probe,
	.remove = gt_hdmi_remove,
	.shutdown = gt_hdmi_shutdown,
	.driver = {
		.name = GT_HDMI_DEVICE_NAME,
		.pm = &gt_hdmi_pm_ops,
	},
	.id_table = s_gt_hdmi_device_id_table,
};

